"""
SSSCTL tests.

:requirement: IDM-SSSD-REQ: Status utility
"""

from __future__ import annotations

import re
import time

import pytest
from pytest_mh.conn import ProcessError
from pytest_mh.conn.ssh import SSHAuthenticationError
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.topology import KnownTopology


@pytest.mark.importance("high")
@pytest.mark.tools
@pytest.mark.ticket(bz=2100789)
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__check_missing_id_provider(client: Client):
    """
    :title: Check id_provider in domain section with sssctl config-check command
    :setup:
        1. Create the sssd.conf, here we are using provider as a LDAP server
    :steps:
        1. Remove id_provider from domain section.
        2. Check error message using sssctl config-check.
    :expectedresults:
        1. Successfully remove id_provider from domain section.
        2. Successfully get the error message.
    :customerscenario: False
    """
    # create sssd.conf and start the sssd, with default configuration with a LDAP server.
    client.sssd.start()

    # remove id_provider parameter from domain section.
    client.sssd.config.remove_option("domain/test", "id_provider")
    client.sssd.config_apply(check_config=False)

    # Check the error message in output of # sssctl config-check
    output = client.host.conn.run("sssctl config-check", raise_on_error=False)
    assert "[rule/sssd_checks]: Attribute 'id_provider' is missing in section 'domain/test'." in output.stdout_lines[1]


@pytest.mark.importance("high")
@pytest.mark.tools
@pytest.mark.ticket(bz=2100789)
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__check_invalid_id_provider(client: Client):
    """
    :title: Check id_provider in domain section with sssctl config-check command with provider
    :setup:
        1. Create the sssd.conf, here we are using provider as a LDAP server
    :steps:
        1. Add invalid, id_provider's value to domain section.
        2. Check error message using sssctl config-check.
    :expectedresults:
        1. Successfully remove id_provider from domain section.
        2. Successfully get the error message.
    :customerscenario: False
    """
    # create sssd.conf and start the sssd, with deafult configuration with a LDAP server.
    client.sssd.start()

    # Add 'invalid' as a id_provider's value in domain section.
    client.sssd.config.remove_option("domain/test", "id_provider")
    client.sssd.domain["id_provider"] = "invalid"
    client.sssd.config_apply(check_config=False)

    # Check the return code of # sssctl config-check command
    output = client.host.conn.run("sssctl config-check", raise_on_error=False)
    assert (
        "[rule/sssd_checks]: Attribute 'id_provider' in section 'domain/test' has an invalid value: invalid"
        in output.stdout_lines[1]
    )


@pytest.mark.ticket(bz=1640576)
@pytest.mark.builtwith("files-provider")
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__user_show_cache_expiration_time(client: Client):
    """
    :title: sssctl user-show reports correct expiration time of local user
    :setup:
        1. Add local users
        2. Configure local domain
        3. Start SSSD
    :steps:
        1. Call sssctl user-show $user
        2. Check correct output
    :expectedresults:
        1. Called successfully
        2. Output is as expected
    :customerscenario: True
    """
    client.local.user("local1").add()
    client.local.user("local2").add()
    client.local.user("local3").add()

    client.sssd.common.local()
    client.sssd.default_domain = "local"
    client.sssd.domain["id_provider"] = "files"
    client.sssd.domain["passwd_files"] = "/etc/passwd"

    client.sssd.start()

    for user in {"local1", "local2", "local3"}:
        cmd = client.sssctl.user_show(user=user)
        assert cmd.rc == 0, "Command call failed"
        assert "Cache entry expiration time: Never" in cmd.stdout, "Wrong output"


@pytest.mark.ticket(bz=1599207)
@pytest.mark.builtwith("files-provider")
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__handle_implicit_domain(client: Client):
    """
    :title: sssctl handle implicit domain
    :setup:
        1. Add local users
        2. Set sssd "enable_files_domain" to "true"
        3. Start SSSD
    :steps:
        1. Call getent passwd user -s sss
        2. Call sssctl user-show --user=$user
        3. Check correct output
    :expectedresults:
        1. Called successfully
        2. Called successfully
        3. Output is correct
    :customerscenario: True
    """
    client.local.user("local1").add()
    client.local.user("local2").add()
    client.local.user("local3").add()

    client.sssd.sssd["enable_files_domain"] = "true"
    client.sssd.start()

    for user in {"local1", "local2", "local3"}:
        assert client.tools.getent.passwd(user, service="sss") is not None
        cmd = client.sssctl.user_show(user=user)
        assert cmd.rc == 0
        assert "Cache entry creation date" in cmd.stdout


@pytest.mark.ticket(bz=1902280)
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__reset_cached_timestamps_to_reflect_changes(client: Client, ldap: LDAP):
    """
    :title: fix sssctl cache-expire to also reset cached timestamp
    :setup:
        1. Add user to LDAP
        2. Add group to LDAP
        3. Set proper domain config options in sssd.conf file
        4. Start SSSD
    :steps:
        1. Call getent group
        2. Modify group entry in LDAP
        3. Call 'sssctl cache-expire -E'
        4. Call getent group
    :expectedresults:
        1. Group is properly cached, user is its member
        2. Member of group is removed, group entry changed
        3. Whole cache is invalidated
        4. User is not member of group anymore
    :customerscenario: True
    """
    u = ldap.user("user1").add()
    ldap.group("group1", rfc2307bis=True).add().add_member(u)

    client.sssd.domain["ldap_schema"] = "rfc2307bis"
    client.sssd.domain["ldap_group_member"] = "member"

    client.sssd.start()

    res1 = client.tools.getent.group("group1")
    assert res1 is not None
    assert "user1" in res1.members

    ldap.group("group1", rfc2307bis=True).remove_member(ldap.user("user1"))
    client.sssctl.cache_expire(everything=True)

    res1 = client.tools.getent.group("group1")
    assert res1 is not None
    assert "user1" not in res1.members


@pytest.mark.importance("high")
@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_invalid_option_name(client: Client):
    """
    :title: sssctl config-check detects mistyped option name
    :setup:
        1. Add wrong_option to domain section
        2. Apply config
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error in config
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.dom("test")["wrong_option"] = "true"
    client.sssd.config_apply(check_config=False)

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config, when SSSD is running"
    assert "Attribute 'wrong_option' is not allowed" in result.stdout, "Wrong error message was returned"


@pytest.mark.importance("high")
@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_missing_domain_name(client: Client):
    """
    :title: sssctl config-check detects mistyped domain name
    :setup:
        1. Create mistyped domain ("domain/")
        2. Start SSSD
    :steps:
        1. Call sssctl config-check, implicitly
        2. Check error message
    :expectedresults:
        1. config-check detects an error in config
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.dom("")["debug_level"] = "9"

    with pytest.raises(ProcessError) as ex:
        client.sssd.start(raise_on_error=True, check_config=True)

    assert ex.match(r"Section \[domain\/\] is not allowed. Check for typos.*"), "Wrong error message was returned"


@pytest.mark.importance("high")
@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_misplaced_option(client: Client):
    """
    :title: sssctl config-check detects misplaced option
    :setup:
        1. In domain set "services" to "nss, pam"
        2. Start SSSD, without config check
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error in config
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.dom("test")["services"] = "nss, pam"

    client.sssd.start(check_config=False)

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"

    pattern = re.compile(r".Attribute 'services' is not allowed in section .*")
    assert pattern.search(result.stdout), "Wrong error message was returned"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_invalid_option_name_in_snippet(client: Client):
    """
    :title: sssctl config-check detects invalid option name in snippet
    :setup:
        1. Create new conf snippet with invalid option name
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error in config snippet
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.fs.write("/etc/sssd/conf.d/01_snippet.conf", "[domain/local]\ninvalid_option = True", mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config snippet"
    assert "Attribute 'invalid_option' is not allowed" in result.stdout, "Wrong error message was returned"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_invalid_section_in_name_in_snippet(client: Client):
    """
    :title: sssctl config-check detects invalid domain name in snippet
    :setup:
        1. Create new conf snippet with invalid domain name
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error in config snippet
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.fs.write("/etc/sssd/conf.d/01_snippet.conf", "[invalid/local]\ninvalid_option = True", mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config snippet"
    assert "Section [invalid/local] is not allowed" in result.stdout, "Wrong error message was returned"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_missing_equal_sign(client: Client):
    """
    :title: sssctl config-check detects missing equals sign
    :setup:
        1. Start SSSD, so default config is automatically created
        2. Edit config file so "=" is missing
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error messages are properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub("id_provider = ", "id_provider ", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Equal sign is missing" in result.stderr, "Wrong error message on stderr"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_special_character_in_option_name(client: Client):
    """
    :title: option name contains special character
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that it contains special character
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub("id_provider", "id_@provider", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Attribute 'id_@provider' is not allowed in section" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_special_character_in_section_name(client: Client):
    """
    :title: section name contains special character
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that it contains special character
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub("domain/", "d$main/", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Section [d$main/local] is not allowed" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_special_character_in_domain_name(client: Client):
    """
    :title: domain name contains special character
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that it contains special character
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub("domain/local", "domain/local@", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Section [domain/local@] is not allowed" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_forward_slash_missing_in_domain_section(client: Client):
    """
    :title: Forward slash is not present between domain name and section name
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that forward slash is missing in section name
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub("domain/local", "domainlocal", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Section [domainlocal] is not allowed" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_invalid_sssd_section_name(client: Client):
    """
    :title: Typo in sssd section name
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that there is typo in sssd section name
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub(".sssd.", "[sssdx]", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Section [sssdx] is not allowed" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_invalid_pam_section_name(client: Client):
    """
    :title: Typo in pam section name
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that there is typo in sssd section name
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub(".pam.", "[pamx]", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Section [pamx] is not allowed" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_invalid_nss_section_name(client: Client):
    """
    :title: Typo in nss section name
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that there is typo in sssd section name
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub(".nss.", "[nssx]", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Section [nssx] is not allowed" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_invalid_permission(client: Client):
    """
    :title: Verify the permission of default configuration file
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Change permission of default config file
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    client.fs.chmod("0777", "/etc/sssd/sssd.conf")
    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "File ownership and permissions check failed" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_missing_closing_bracket(client: Client):
    """
    :title: Missing closing bracket in sssd section name
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that there is missing closing bracket
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub(".nss.", "[nssx", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "No closing bracket" in result.stderr, "Wrong error message on stderr"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_missing_opening_bracket(client: Client):
    """
    :title: Missing opening bracket in domain name
    :setup:
        1. Start SSSD, so default config is autimatically created
        2. Edit config file in a way that there is missing opening bracket
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
        1. Is there no "No opening bracket" error like the closing bracket one?
    """
    client.sssd.common.local()
    client.sssd.start()
    conf = re.sub(".domain/local.", "domain/local]", client.fs.read("/etc/sssd/sssd.conf"))
    client.fs.write("/etc/sssd/sssd.conf", conf, mode="600")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "Equal sign is missing" in result.stderr, "Wrong error message on stderr"
    assert "Failed to parse configuration" in result.stderr, "Wrong error message on stderr"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_invalid_options_in_two_domains(client: Client):
    """
    :title: Verify typo in option name with multiple domains in default configuration file
    :setup:
        1. Configure two ldap domains
        2. Make typo in both domains
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error messages are properly set
    :customerscenario: False
    : with other invalid option tests?
    """
    client.sssd.common.local()
    client.sssd.dom("ldap1")["ldap_ri"] = "ldaps://invalid"
    client.sssd.dom("ldap1")["id_provider"] = "ldap"
    client.sssd.dom("ldap2")["ldap_ri"] = "ldaps://invalid"
    client.sssd.dom("ldap2")["id_provider"] = "ldap"
    client.sssd.sssd["domains"] = "ldap1, ldap2"
    client.sssd.config_apply(check_config=False)

    res = client.sssctl.config_check()
    assert res.rc != 0, "Config-check did not detect misconfigured config"
    assert "Issues identified by validators: 2" in res.stdout, "Wrong number of issues found by validators"
    assert "Attribute 'ldap_ri' is not allowed in section 'domain/ldap1'" in res.stdout, "Wrong error message"
    assert "Attribute 'ldap_ri' is not allowed in section 'domain/ldap2'" in res.stdout, "Wrong error message"


@pytest.mark.tools
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_config_does_not_exist(client: Client):
    """
    :title: sssctl detects missing config
    :setup:
        1. Start SSSD, so default config is automatically created
        2. Remove config
    :steps:
        1. Call sssctl config-check
        2. Check error message
    :expectedresults:
        1. config-check detects an error
        2. Error message is properly set
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.start()
    client.fs.rm("/etc/sssd/sssd.conf")

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert "File /etc/sssd/sssd.conf does not exist" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.ticket(bz=1677994)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_ldap_host_object_class_in_domain(client: Client):
    """
    :title: sssctl config-check allow ldap_host_object_class in domain section
    :setup:
        1. Add ldap_host_object_class to local domain section
        2. Start SSSD
    :steps:
        1. Call sssctl config-check
    :expectedresults:
        1. config-check succeed
    :customerscenario: True
    """
    client.sssd.default_domain = "local"
    client.sssd.common.local()
    client.sssd.domain["ldap_host_object_class"] = "ipService"
    client.sssd.start(check_config=False)

    result = client.sssctl.config_check()
    assert result.rc == 0, "Config-check failed"


@pytest.mark.tools
@pytest.mark.ticket(bz=1677994)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_ldap_host_object_class_not_allowed_in_sssd(client: Client):
    """
    :title: sssctl config-check do not allow ldap_host_object_class in sssd section
    :setup:
        1. Add ldap_host_object_class to sssd section
        2. Start SSSD
    :steps:
        1. Call sssctl config-check
    :expectedresults:
        1. config-check succeed
    :customerscenario: True
    """
    client.sssd.common.local()
    client.sssd.sssd["ldap_host_object_class"] = "ipService"
    client.sssd.start(check_config=False)

    result = client.sssctl.config_check()
    assert result.rc != 0, "Config-check did not detect misconfigured config"
    assert (
        "Attribute 'ldap_host_object_class' is not allowed in section 'sssd'" in result.stdout
    ), "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.ticket(bz=1856861)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_enabling_2fa_prompting(client: Client):
    """
    :title: False warnings are logged in sssd.log file after enabling 2FA prompting
    :setup:
        1. Enable Two factor authentication
        2. Start SSSD
    :steps:
        1. Call sssctl config-check
    :expectedresults:
        1. config-check succeed
    :customerscenario: True
    """
    client.sssd.common.local()
    client.sssd.section("prompting/2fa/sshd")["first_prompt"] = "Enter OTP Token Value:"
    client.sssd.section("prompting/2fa/sshd")["single_prompt"] = "True"

    client.sssd.section("prompting/2fa")["first_prompt"] = "prompt1"
    client.sssd.section("prompting/2fa")["second_prompt"] = "prompt2"
    client.sssd.section("prompting/2fa")["single_prompt"] = "True"

    client.sssd.start(check_config=False)
    result = client.sssctl.config_check()
    assert result.rc == 0, "Config-check failed"


@pytest.mark.tools
@pytest.mark.ticket(bz=1791892)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_auto_private_groups_in_child_domains(client: Client):
    """
    :title: sssctl config-check detects false positive when auto_private_groups is enabled or disabled in child domains
    :setup:
        1. Enable auto_private_groups in child domain
        2. Disable auto_private_groups in second child domain
        3. Start SSSD
    :steps:
        1. Call sssctl config-check
    :expectedresults:
        1. config-check succeed
    :customerscenario: True
    """
    client.sssd.common.local()
    client.sssd.sssd["domains"] = "td5f4f77.com"
    client.sssd.subdom("td5f4f77.com", "one5f4f77.td5f4f77.com")["auto_private_groups"] = "True"
    client.sssd.subdom("td5f4f77.com", "two5f4f77.td5f4f77.com")["auto_private_groups"] = "False"

    client.sssd.start(check_config=False, debug_level=None)
    result = client.sssctl.config_check()
    assert result.rc == 0, "Config-check failed"


@pytest.mark.tools
@pytest.mark.ticket(bz=1723273)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_non_default_config_location_missing_snippet_directory(client: Client):
    """
    :title: sssctl config-check complains about non existing snippet directory when config is non default
    :setup:
        1. Copy sssd.conf file to different directory
    :steps:
        1. Call sssctl config-check on that different directory
        2. Check error message
    :expectedresults:
        1. config-check failed
        2. Error message is properly set
    :customerscenario: True
    """
    client.sssd.common.local()
    client.sssd.config_apply()
    client.fs.mkdir("/tmp/test/")
    client.fs.copy("/etc/sssd/sssd.conf", "/tmp/test/")

    result = client.sssctl.config_check(config="/tmp/test/sssd.conf")
    assert result.rc != 0, "Config-check successfully finished"
    assert "Directory /tmp/test/conf.d does not exist" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.ticket(bz=1723273)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_non_default_config_location_invalid_permission(client: Client):
    """
    :title: sssctl config-check complains about proper permission when config is non default
    :setup:
        1. Copy sssd.conf file to different directory and set it wrong permission
    :steps:
        1. Call sssctl config-check on that different directory
        2. Check error message
    :expectedresults:
        1. config-check failed
        2. Error message is properly set
    :customerscenario: True
    """
    client.sssd.common.local()
    client.sssd.config_apply()
    client.fs.mkdir("/tmp/test/")
    client.fs.copy("/etc/sssd/sssd.conf", "/tmp/test/sssd.conf", mode="777")

    result = client.sssctl.config_check(config="/tmp/test/sssd.conf")
    assert result.rc != 0, "Config-check successfully finished"
    assert "File ownership and permissions check failed" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.ticket(bz=1723273)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_non_default_config_location_invalid_option_name(client: Client):
    """
    :title: sssctl config-check detects typo in option name when config is non default
    :setup:
        1. Copy sssd.conf file to different directory and mistype option name
    :steps:
        1. Call sssctl config-check on that different directory
        2. Check error message
    :expectedresults:
        1. config-check failed
        2. Error message is properly set
    :customerscenario: True
    """
    client.sssd.common.local()
    client.sssd.default_domain = "local"
    client.sssd.domain["search_base"] = "True"
    client.sssd.config_apply(check_config=False)

    client.fs.mkdir("/tmp/test/")
    client.fs.copy("/etc/sssd/sssd.conf", "/tmp/test/")

    result = client.sssctl.config_check(config="/tmp/test/sssd.conf")
    assert result.rc != 0, "Config-check successfully finished"
    assert (
        "Attribute 'search_base' is not allowed in section 'domain/local'" in result.stdout
    ), "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.ticket(bz=1723273)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_non_default_config_location_with_snippet_directory(client: Client):
    """
    :title: sssctl config-check does not complain about missing snippet directory after adding with proper permission
    :setup:
        1. Copy sssd.conf file to different directory and create conf.d directory
    :steps:
        1. Call sssctl config-check on that different directory
        2. Check error message
    :expectedresults:
        1. config-check failed
        2. Error message is properly set
    :customerscenario: True
    """
    client.sssd.common.local()
    client.sssd.config_apply(check_config=False)

    client.fs.mkdir("/tmp/test/")
    client.fs.mkdir("/tmp/test/conf.d", mode="700")
    client.fs.copy("/etc/sssd/sssd.conf", "/tmp/test/")

    result = client.sssctl.config_check(config="/tmp/test/sssd.conf")
    assert result.rc == 0, "Config-check failed"
    assert "Directory /tmp/test/conf.d does not exist" not in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.ticket(bz=1723273)
@pytest.mark.topology(KnownTopology.Client)
def test_sssctl__check_non_existing_snippet(client: Client):
    """
    :title: sssctl config-check detects non existing snippet directory
    :setup:
        1. Start SSSD, so default config is autimatically created
    :steps:
        1. Call sssctl config-check with non existing snippet
        2. Check error message
    :expectedresults:
        1. config-check failed
        2. Error message is properly set
    :customerscenario: True
    """
    client.sssd.common.local()
    client.sssd.start()
    result = client.sssctl.config_check(snippet="/does/not/exist")
    assert result.rc != 0, "Config-check successfully finished"
    assert "Directory /does/not/exist does not exist" in result.stdout, "Wrong error message on stdout"


@pytest.mark.tools
@pytest.mark.ticket(bz=1294670)
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__analyze_list(client: Client, ldap: LDAP):
    """
    :title: "sssctl analyze request list" show captured nss related requests from sssd log
    :setup:
        1. Add user and group
        2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
        3. Start SSSD
    :steps:
        1. Call id user1 and getent group group1
        2. Call sssctl analyze request list, also with -v
        3. Find "getent" and " id" in result
        4. Clear cache
        5. Call getent passwd user
        6. Call sssctl analyze request list, also with -v
        7. Find "CID #1" and "getent" in result
    :expectedresults:
        1. Called successfully, information is stored in logs
        2. Called successfully
        3. Strings found
        4. Cache cleared
        5. Called successfully
        6. Called successfully
        7. Strings found
    :customerscenario: True
    """
    ldap.user("user1").add()
    ldap.group("group1").add()
    client.sssd.nss["debug_level"] = "9"
    client.sssd.pam["debug_level"] = "9"
    client.sssd.domain["debug_level"] = "9"
    client.sssd.start()

    assert client.tools.getent.group("group1"), "getent group1 failed"
    assert client.tools.id("user1"), "id user1 failed"

    res = client.sssctl.analyze_request("list")
    assert res.rc == 0, "sssctl analyze call failed"
    assert "getent" in res.stdout, "'getent' not found in analyze list output"
    assert " id" in res.stdout or "coreutils" in res.stdout, "' id' or 'coreutils' not found in analyze list output"
    res = client.sssctl.analyze_request("list -v")
    assert res.rc == 0, "sssctl analyze call failed"
    assert "getent" in res.stdout, "'getent' not found in analyze list -v output"
    assert " id" in res.stdout or "coreutils" in res.stdout, "' id' or 'coreutils' not found in analyze list -v output"

    client.sssd.stop()
    client.sssd.clear(db=True, memcache=True, logs=True)
    client.sssd.start()

    assert client.tools.getent.passwd("user1")
    res = client.sssctl.analyze_request("list")
    assert res.rc == 0, "sssctl analyze call failed"
    assert "CID #1" in res.stdout, "CID #1 not found in analyze list -v output"
    assert "getent" in res.stdout, "getent not found in analyze list -v output"
    res = client.sssctl.analyze_request("list -v")
    assert res.rc == 0, "sssctl analyze call failed"
    assert "CID #1" in res.stdout, "CID #1 not found in analyze list -v output"
    assert "getent" in res.stdout, "getent not found in analyze list -v output"


@pytest.mark.tools
@pytest.mark.ticket(bz=1294670, gh=6298)
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__analyze_non_default_log_location(client: Client, ldap: LDAP):
    """
    :title: "sssctl analyze" parse sssd logs from non-default location when SSSD is not running
    :setup:
        1. Add user
        2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
        3. Start SSSD
    :steps:
        1. Call id user1 and login user via ssh
        2. Copy sssd logs to diferent location
        3. Stop sssd and remove config, logs and cache
        4. sssctl analyze --logdir PATH parse logs from PATH location
    :expectedresults:
        1. Information is stored in logs
        2. Copied successfully
        3. Stopped and cleared successfully
        4. Output is correct
    :customerscenario: True
    """
    ldap.user("user1").add(password="Secret123")
    client.sssd.nss["debug_level"] = "9"
    client.sssd.pam["debug_level"] = "9"
    client.sssd.domain["debug_level"] = "9"
    client.sssd.start()

    assert client.tools.id("user1@test"), "call 'id user1@test' failed"
    client.ssh("user1", "Secret123").connect()

    client.fs.copy("/var/log/sssd", "/tmp/copy/")
    client.sssd.stop()
    client.sssd.clear(config=True, logs=True)

    res = client.sssctl.analyze_request(command="show 1 --pam", logdir="/tmp/copy/")
    assert "SSS_PAM_AUTHENTICATE" in res.stdout
    assert "SSS_PAM_ACCT_MGMT" in res.stdout
    assert "SSS_PAM_SETCRED" in res.stdout

    res = client.sssctl.analyze_request(command="list", logdir="/tmp/copy/")
    assert " id" in res.stdout or "coreutils" in res.stdout, "' id' or 'coreutils' not found in analyze list output"
    assert "sshd" in res.stdout or "coreutils" in res.stdout, "sshd or coreutils not found in output"

    res = client.sssctl.analyze_request(command="list -v", logdir="/tmp/copy/")
    assert " id" in res.stdout or "coreutils" in res.stdout, "' id' or 'coreutils' not found in analyze list -v output"
    assert "sshd" in res.stdout or "coreutils" in res.stdout, "sshd or coreutils not found in output"


@pytest.mark.tools
@pytest.mark.ticket(bz=1294670)
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__analyze_pam_logs(client: Client, ldap: LDAP):
    """
    :title: "sssctl analyze" to parse pam authentication requests from logs
    :setup:
        1. Add user
        2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
        3. Start SSSD
        4. Log in as user via ssh
    :steps:
        1. sssctl analyze with --pam option
        2. Result of command is login related
    :expectedresults:
        1. Called successfully
        2. Output is login related
    :customerscenario: True
    """
    ldap.user("user1").add()
    client.sssd.nss["debug_level"] = "9"
    client.sssd.pam["debug_level"] = "9"
    client.sssd.start()

    client.ssh("user1", "Secret123").connect()

    result = client.sssctl.analyze_request("show 1 --pam")
    assert result.rc == 0
    assert "CID #1" in result.stdout

    assert "SSS_PAM_AUTHENTICATE" in result.stdout
    assert "SSS_PAM_ACCT_MGMT" in result.stdout
    assert "SSS_PAM_SETCRED" in result.stdout


@pytest.mark.tools
@pytest.mark.ticket(bz=2013259)
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__analyze_tevent_id(client: Client, ldap: LDAP):
    """
    :title: "sssctl analyze" to parse tevent chain IDs from logs
    :setup:
        1. Add user
        2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
        3. Start SSSD
        4. Log in as user via ssh
    :steps:
        1. Call sssctl analyze request show 1 --pam
        2. Confirm tevent chain IDs(RID) is showing in logs
    :expectedresults:
        1. Called successfully
        2. Output is correct
    :customerscenario: True
    """
    ldap.user("user1").add()
    client.sssd.nss["debug_level"] = "9"
    client.sssd.pam["debug_level"] = "9"
    client.sssd.domain["debug_level"] = "9"
    client.sssd.start()

    client.ssh("user1", "Secret123").connect()

    result = client.sssctl.analyze_request("show 1 --pam")
    assert result.rc == 0
    assert "RID#" in result.stdout, "RID# was not found in the output"
    assert "user1@test" in result.stdout, "user1@test was not found in the output"


@pytest.mark.tools
@pytest.mark.ticket(bz=2013260)
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__analyze_child_logs(client: Client, ldap: LDAP):
    """
    :title: "sssctl analyze" to parse child logs
    :setup:
        1. Add user
        2. Enable debug_level to 9 in the 'nss', 'pam' and domain section
        3. Start SSSD
    :steps:
        1. Log in as user via ssh
        2. Call sssctl analyze to check logs
        3. Clear cache and restart SSSD
        4. Log in as user via ssh with wrong password
        5. Call sssctl analyze to check logs
    :expectedresults:
        1. Logged in successfully
        2. Logs contain login related logs
        3. Succesfully
        4. Failed to login
        5. Logs contain info about failed login
    :customerscenario: True
    """
    ldap.user("user1").add()
    client.sssd.nss["debug_level"] = "9"
    client.sssd.pam["debug_level"] = "9"
    client.sssd.domain["debug_level"] = "9"
    client.sssd.start()

    client.ssh("user1", "Secret123").connect()

    result = client.sssctl.analyze_request("show --pam --child 1")
    assert result.rc == 0
    assert "user1@test" in result.stdout
    assert "SSS_PAM_AUTHENTICATE" in result.stdout

    client.sssd.stop()
    client.sssd.clear(db=True, memcache=True, logs=True)
    client.sssd.start()
    time.sleep(5)

    with pytest.raises(SSHAuthenticationError):
        client.ssh("user1", "Wrong").connect()
    result = client.sssctl.analyze_request("show --pam --child 1")
    assert (
        "Authentication failure to the client" in result.stdout
    ), "'Authentication failure to the client' was not found"


@pytest.mark.tools
@pytest.mark.ticket(bz=[2142960, 2142794, 2142961])
@pytest.mark.topology(KnownTopology.LDAP)
def test_sssctl__analyze_without_root_privileges(client: Client, ldap: LDAP):
    """
    :title: "sssctl analyze" command does not require "root" privileges
    :setup:
        1. Add user with proper password
        2. Start SSSD
        3. Fetch information about user
        4. Copy logs to different location
        5. Change ownership of copied file to user
    :steps:
        1. Call "sssctl analyze --logdir /tmp/copy request show 1" as root
        2. Call "sssctl analyze --logdir /tmp/copy request show 1" as user
        3. Check that outputs match
        4. Username is stored in outputs
    :expectedresults:
        1. Called successfully
        2. Called successfully
        3. Outputs are the same
        4. Username is stored in outputs
    :customerscenario: True
    """
    ldap.user("user1").add(password="Secret123")
    client.sssd.start()
    client.tools.id("user1")
    client.fs.copy("/var/log/sssd", "/tmp/copy/")
    client.fs.chown("/tmp/copy", "user1", args=["--recursive"])

    result_root = client.sssctl.analyze_request(command="show 1", logdir="/tmp/copy")
    result_user = client.ssh("user1", "Secret123").run("sssctl analyze --logdir /tmp/copy request show 1")
    assert result_root.rc == 0, "sssctl analyze call failed as root"
    assert result_user.rc == 0, "sssctl analyze call failed as user1"
    assert result_root.stdout == result_user.stdout, "the outputs are different"
    assert "user1" in result_user.stdout, "user1 is not in the outputs"



""" Tests for Memcache for SIDs

:requirement: Memory cache for SID
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

import pytest
from sssd.testlib.common.utils import sssdTools


@pytest.mark.tier1_3
@pytest.mark.memcachesid
class Testmemcachesid(object):
    def test_0001_memcache_sid(self, multihost, adjoin, create_aduser_group):
        """
        :title: Verify memcache for SID
        :id: f7fce9c5-5ba6-428b-8e9b-5e07a88b5050
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1245367
        :customerscenario: true
        :steps:
          1. get uid of user
          2. get gid of group
          3. Clear sssd cache
          4. Lookup sid-by-uid with pysss_nss_idmap
          5. Lookup sid-by-uid with pysss_nss_idmap again
          6. Lookup sid-by-gid with pysss_nss_idmap
          7. Lookup sid-by-gid with pysss_nss_idmap again
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Lookup should run through nss responder
          5. Lookup should read from memory cache of SID
          6. Lookup should run through nss responder
          7. Lookup should read from memory cache of SID
        """
        req_pkg = 'yum install -y strace python3-libsss_nss_idmap'
        adjoin(membersw='adcli')
        multihost.client[0].run_command(req_pkg)
        ad_domain = multihost.ad[0].domainname
        # Create AD user and group
        (aduser, adgroup) = create_aduser_group
        client = sssdTools(multihost.client[0], multihost.ad[0])
        dom_section = f'domain/{client.get_domain_section_name()}'
        sssd_params = {
            'ad_domain': multihost.ad[0].domainname.upper(),
            'debug_level': '9',
            'use_fully_qualified_names': 'True',
            'cache_credentials': 'True',
        }
        client.sssd_conf(dom_section, sssd_params)
        client.clear_sssd_cache()
        lookup_cmd_user = f'id -u {aduser}@{ad_domain}'
        cmd = multihost.client[0].run_command(lookup_cmd_user)
        ad_uid = cmd.stdout_text.rstrip()
        lookup_cmd_group = f'getent group {adgroup}@{ad_domain} | cut -f3 -d:'
        cmd = multihost.client[0].run_command(lookup_cmd_group)
        ad_gid = cmd.stdout_text.rstrip()
        file_dict = {'getsidbyuid': ad_uid, 'getsidbygid': ad_gid}
        for k, v in file_dict.items():
            run_args = f'python3 -c "import pysss_nss_idmap;pysss_nss_idmap.{k}({v})"'

            for i in 'before', 'after':
                strace_file = f'/opt/{k}_{i}.trace'
                cmd = f'strace -fxvto {strace_file} {run_args}'
                multihost.client[0].run_command(cmd)
                cmd = f'grep /var/lib/sss/pipes/nss {strace_file}'
                chk_log = multihost.client[0].run_command(cmd, raiseonerr=False)
                rm_cmd = f'rm -f {strace_file}'
                multihost.client[0].run_command(rm_cmd)
                if i == 'before':
                    assert chk_log.returncode == 0
                if i == 'after':
                    assert chk_log.returncode == 1



import argparse


# Based on patch from https://bugs.python.org/issue9341
class SubparsersAction(argparse._SubParsersAction):
    """
    Provide a subparser action that can create subparsers with ability of
    grouping arguments.

    It is based on the patch from:

        - https://bugs.python.org/issue9341
    """

    class _PseudoGroup(argparse.Action):
        def __init__(self, container, title):
            super().__init__(option_strings=[], dest=title)
            self.container = container
            self._choices_actions = []

        def add_parser(self, name, **kwargs):
            # add the parser to the main Action, but move the pseudo action
            # in the group's own list
            parser = self.container.add_parser(name, **kwargs)
            choice_action = self.container._choices_actions.pop()
            self._choices_actions.append(choice_action)
            return parser

        def _get_subactions(self):
            return self._choices_actions

        def add_parser_group(self, title):
            # the formatter can handle recursive subgroups
            grp = SubparsersAction._PseudoGroup(self, title)
            self._choices_actions.append(grp)
            return grp

    def add_parser_group(self, title):
        """
        Add new parser group.

        :param title: Title.
        :type title: str
        :return: Parser group that can have additional parsers attached.
        :rtype: ``argparse.Action`` extended with ``add_parser`` method
        """
        grp = self._PseudoGroup(self, title)
        self._choices_actions.append(grp)
        return grp


class Option:
    """
    Group option attributes for command/subcommand options
    """
    def __init__(self, name, help_msg, opt_type, short_opt=None):
        self.name = name
        self.short_opt = short_opt
        self.help_msg = help_msg
        self.opt_type = opt_type



"""Automation of kcm related bugs

:requirement: IDM-SSSD-REQ :: SSSD KCM as default Kerberos CCACHE provider
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

from __future__ import print_function
import re
import pytest
import time
from sssd.testlib.common.ssh2_python import check_login_client
from sssd.testlib.common.utils import sssdTools
from constants import ds_instance_name


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups')
@pytest.mark.kcm
class TestKcm(object):
    """
    This is for kcm bugs automation
    """
    @pytest.mark.tier1_2
    def test_client_timeout(self, multihost, backupsssdconf):
        """
        :title: kcm: Increase client idle
         timeout to 5 minutes
        :id: 6933cb85-1616-4b7f-a049-e81ab4c05347
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1884205
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'debug_level': '9'}
        client.sssd_conf('kcm', domain_params)
        multihost.client[0].service_sssd('restart')
        user = 'foo1@example1'
        check_login_client(multihost, user, 'Secret123')
        sssdTools(multihost.client[0]).clear_sssd_cache()
        multihost.client[0].run_command("systemctl restart sssd-kcm")
        multihost.client[0].run_command("> /var/log/sssd/sssd_kcm.log")
        start_time = time.time()
        multihost.client[0].run_command("kinit foo1 <&- & ")
        end_time = time.time()
        assert end_time - start_time >= 300
        grep_cmd = multihost.client[0].run_command("grep"
                                                   " 'Terminated"
                                                   " client'"
                                                   " /var/log/sssd/"
                                                   "sssd_kcm.log")
        assert 'Terminated client' in grep_cmd.stdout_text

    @pytest.mark.tier1_2
    def test_refresh_contain_timestamp(self,
                                       multihost,
                                       backupsssdconf):
        """
        :title: kcm: First smart refresh query contains
         modifyTimestamp even if the modifyTimestamp is 0
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1926454
        :customerscenario: true
        :id: 09f654c4-759d-11eb-bfff-002b677efe14
        :steps:
          1. Configure SSSD with sudo
          2. Leave ou=sudoers empty - do not define any rules
          3. See that smart refresh does not contain
             modifyTimestamp in the filter
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        ldap_params = {'domains': 'example1'}
        tools.sssd_conf('sssd', ldap_params)
        ldap_params = {'sudo_provider': 'ldap',
                       'ldap_sudo_smart_refresh_interval': '60'}
        tools.sssd_conf('domain/%s' % (ds_instance_name), ldap_params)
        multihost.client[0].service_sssd('restart')
        multihost.client[0].run_command("> /var/log/sssd/sssd_example1.log")
        time.sleep(65)
        log_location = "/var/log/sssd/sssd_example1.log"
        grep_cmd = multihost.client[0].run_command(f"grep "
                                                   f"'calling "
                                                   f"ldap_search_ext with' "
                                                   f"{log_location}")
        assert 'modifyTimestamp>=' not in grep_cmd.stdout_text

    @pytest.mark.tier1_2
    def test_kcm_check_socket_path(self, multihost, backupsssdconf):
        """
        :title: kcm: Test socket path when sssd-kcm is activated by systemd
        :id: 6425bf2c-d07e-4d65-b15d-946141422f96
        :ticket: https://github.com/SSSD/sssd/issues/5406
        """
        # Start from a known-good state after removing log file and adding a
        # new socket path
        client = sssdTools(multihost.client[0])
        domain_log = '/var/log/sssd/sssd_kcm.log'
        multihost.client[0].service_sssd('stop')
        client.service_ctrl('stop', 'sssd-kcm')
        client.remove_sss_cache(domain_log)
        domain_params = {'debug_level': '9',
                         'socket_path': '/some_path/kcm.socket'}
        client.sssd_conf('kcm', domain_params)
        multihost.client[0].service_sssd('start')
        # After starting sssd-kcm, latest sssd_kcm.log will generate
        client.service_ctrl('start', 'sssd-kcm')
        # Give sssd some time to load
        time.sleep(2)
        # Check log file for the expected warning message
        log = multihost.client[0].get_file_contents(domain_log).decode('utf-8')
        msg = "Warning: socket path defined in systemd unit .*." \
              "and.sssd.conf...some_path.kcm.socket..don't.match"
        find = re.compile(r'%s' % msg)
        assert find.search(log)

    @pytest.mark.tier1_2
    def test_expired_tickets(self, multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: kcm_provider: Expired tickets are removed
        :id: db532785-a00f-4be9-b413-592e0550fe9c
        :requirement: sssd-kcm does not appear to expire Kerberos tickets (RFE: sssd_kcm
          should have the option to automatically delete the expired tickets)
        :setup:
         1. Configure short ticket lifetime in krb5.
         2. Create a user.
         3. Cleanup all ticket caches
        :steps:
          1. Create about 64 tickets for user
          2. Try to create 65th ticket
          3. Wait for them to expire
          4. Try to create about 10 more tickets
          5. Check the sssd kcm log
        :expectedresults:
          1. Tickets are created
          2. Ticket is not created
          3. Tickets are expired
          4. New tickets are created
          5. Log contains a message about deleting expired credentials
        :customerscenario: True
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1900973
        """
        client = sssdTools(multihost.client[0])
        client.sssd_conf('kcm', {'debug_level': '9'})
        multihost.client[0].service_sssd('restart')

        sssdTools(multihost.client[0]).clear_sssd_cache()
        # Setup short ticker validity (60s)  in krb5.conf using ticket_lifetime
        backup_krb5 = 'cp -rf /etc/krb5.conf /etc/krb5.conf.bak'
        restore_krb5 = 'mv /etc/krb5.conf.bak /etc/krb5.conf ; ' \
                       'restorecon -Rv /etc/krb5.conf'
        edit_krb5_conf = 'sed -i "s/ticket_lifetime.*/ticket_lifetime = ' \
            '60/" /etc/krb5.conf'
        multihost.client[0].run_command(backup_krb5, raiseonerr=False)
        multihost.client[0].run_command(edit_krb5_conf, raiseonerr=False)
        multihost.client[0].run_command("systemctl restart sssd-kcm")
        multihost.client[0].run_command("> /var/log/sssd/sssd_kcm.log")
        multihost.client[0].run_command("kdestroy -A", raiseonerr=False)
        for i in range(1, 65):
            multihost.client[0].run_command(
                f"kinit -c KCM:0:12345{i} foo1",
                stdin_text="Secret123",
                raiseonerr=False
            )
        # This credential should not be created due to secrets being full now.
        cmd_fail = multihost.client[0].run_command(
            "kinit -c KCM:0:666666 foo1",
            stdin_text="Secret123",
            raiseonerr=False
        )
        log_str_1 = multihost.client[0].get_file_contents(
            "/var/log/sssd/sssd_kcm.log").decode('utf-8')
        # Wait for secrets to expire.
        time.sleep(120)
        multihost.client[0].run_command("date; klist -l", raiseonerr=False)
        fail_count = 0
        for i in range(65, 75):
            cmd = multihost.client[0].run_command(
                f"kinit -c KCM:0:12345{i} foo1",
                stdin_text="Secret123",
                raiseonerr=False
            )
            fail_count += cmd.returncode
        log_str = multihost.client[0].get_file_contents(
            "/var/log/sssd/sssd_kcm.log").decode('utf-8')
        multihost.client[0].run_command(restore_krb5, raiseonerr=False)
        multihost.client[0].run_command("kdestroy -A", raiseonerr=False)
        assert fail_count == 0, f"At least one kinit failed. Failures: {fail_count}."
        assert cmd_fail.returncode != 0, "kinit succeeded but should have failed."
        assert "The maximum number of stored secrets has been reached" in log_str_1
        assert "Removing the oldest expired credential" in log_str

    @pytest.mark.tier1_2
    @staticmethod
    def test_kcm_logrotate(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: sssd_kcm.log is rotated with other sssd logs
        :id: 9ac9a11c-c176-431b-b690-03181f014c25
        :setup:
         1. Create a user.
        :steps:
          1. Create a ticket for a user and collect kcm log file
          2. Initiate logrotate for sssd and  collect the sssd kcm log
          3. Swich off sssd-kcm service and initiate logrotate
          4. Swich off sssd service and initiate logrotate
        :expectedresults:
          1. Ticket is created
          2. Log file size is lower that in step 1.
          3. Logrotate does not fail.
          4. Logrotate does not fail.
        :customerscenario: True
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=2176768
        """
        client = sssdTools(multihost.client[0])
        client.sssd_conf('kcm', {'debug_level': '9'})
        sssdTools(multihost.client[0]).clear_sssd_cache()
        client.service_ctrl('restart', 'sssd-kcm')
        kinit = multihost.client[0].run_command(
            "kinit -c KCM:0:666666 foo1",
            stdin_text="Secret123",
            raiseonerr=False
        )
        log_str_len_start = len(multihost.client[0].get_file_contents(
            "/var/log/sssd/sssd_kcm.log").decode('utf-8'))
        multihost.client[0].run_command("logrotate -f /etc/logrotate.d/sssd")
        time.sleep(20)
        log_str_len_end = len(multihost.client[0].get_file_contents(
            "/var/log/sssd/sssd_kcm.log").decode('utf-8'))
        client.service_ctrl('stop', 'sssd-kcm')
        logrotate_kcm = multihost.client[0].run_command("logrotate -f /etc/logrotate.d/sssd")
        multihost.client[0].service_sssd('stop')
        logrotate_sssd = multihost.client[0].run_command("logrotate -f /etc/logrotate.d/sssd")
        assert kinit.returncode == 0, "kinit failed."
        assert log_str_len_start > log_str_len_end, "Log was not rotated!"
        assert logrotate_kcm.returncode == 0, "Logrotation error when sssd-kcm is stopped!"
        assert logrotate_sssd.returncode == 0, "Logrotation error when sssd is stopped!"



from __future__ import print_function
import pytest
from sssd.testlib.common.utils import sssdTools


@pytest.mark.adsites
class Testadsites(object):
    """
    @Title: IDM-SSSD-TC: ad_provider: adsites:
    Improve AD site discovery process
    Test cases for BZ: 1819012

    @Steps:
    1. Join client to AD
    2. Start SSSD and enable debug
    3. Create secondary site, move second domain controller to second site
    """
    @pytest.mark.adsites
    @pytest.mark.c_ares
    def test_001_ad_startup_discovery(self, multihost, adjoin):
        """
        @Title: IDM-SSSD-TC: ad_startup_discovery
        * grep sssd domain logs for cldap ping
        * grep sssd logs for cldap ping parallel batch
        * grep sssd logs for cldap ping domain discovery
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = client.get_domain_section_name()
        domain_section = 'domain/{}'.format(domain)
        sssd_params = {'debug_level': '0xFFF0'}
        client.sssd_conf(domain_section, sssd_params)

        ad1 = multihost.ad[0].hostname
        ad2 = multihost.ad[1].hostname
        multihost.client[0].service_sssd('start')

        cmd_id = 'id Administrator@%s' % domain
        multihost.client[0].run_command(cmd_id)

        cmd_check_ping = 'grep -ire ad_cldap_ping_send ' \
                         '/var/log/sssd/sssd_%s.log | ' \
                         'grep -ire \"Found 2 domain controllers in domain ' \
                         'Default-First-Site-Name._sites.%s\"'\
                         % (domain, domain)
        check_ping = multihost.client[0].run_command(cmd_check_ping,
                                                     raiseonerr=False)
        assert check_ping.returncode == 0
        cmd_check_batch1 = 'grep -ire ad_cldap_ping_parallel_batch' \
                           ' /var/log/sssd/sssd_%s.log | ' \
                           'grep -ire \" %s\"' % (domain, ad1)
        check_batch1 = multihost.client[0].run_command(cmd_check_batch1,
                                                       raiseonerr=False)
        cmd_check_batch2 = 'grep -ire ad_cldap_ping_parallel_batch' \
                           ' /var/log/sssd/sssd_%s.log | ' \
                           'grep -ire \" %s\"' % (domain, ad2)
        check_batch2 = multihost.client[0].run_command(cmd_check_batch2,
                                                       raiseonerr=False)
        if check_batch1.returncode == 0 or check_batch2.returncode == 0:
            assert True
        else:
            assert False
        cmd_check_discovery = 'grep -ire ad_cldap_ping_domain_discovery_done' \
                              ' /var/log/sssd/sssd_%s.log | ' \
                              'grep -ire \"Found 2 domain controllers in' \
                              ' domain Default-First-Site-Name._sites.%s\"'\
                              % (domain, domain)
        check_discovery = multihost.client[0].run_command(cmd_check_discovery,
                                                          raiseonerr=False)
        assert check_discovery.returncode == 0

    @pytest.mark.adsites
    @pytest.mark.c_ares
    def test_002_ad_startup_discovery_one_server_unreachable(self, multihost,
                                                             adjoin):
        """
        @Title: IDM-SSSD-TC: ad_startup_discovery_one_server_unreachable
        * grep sssd domain logs for cldap ping
        * grep sssd logs for cldap ping parallel batch
        * grep sssd logs for cldap ping domain discovery
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = client.get_domain_section_name()
        domain_section = 'domain/{}'.format(domain)
        sssd_params = {'debug_level': '0xFFF0'}
        client.sssd_conf(domain_section, sssd_params)

        ad1 = multihost.ad[0].hostname
        ad2 = multihost.ad[1].hostname
        ad2ip = multihost.ad[1].ip

        cmd_dnf_firewalld = 'dnf install -y firewalld'
        multihost.client[0].run_command(cmd_dnf_firewalld)
        cmd_start_firewalld = 'systemctl start firewalld'
        multihost.client[0].run_command(cmd_start_firewalld)
        fw_add = 'firewall-cmd --permanent --direct --add-rule ipv4 ' \
                 'filter OUTPUT 0 -d %s -j DROP' % ad2ip
        fw_reload = 'firewall-cmd --reload'
        multihost.client[0].run_command(fw_add, raiseonerr=True)
        multihost.client[0].run_command(fw_reload, raiseonerr=True)
        multihost.client[0].service_sssd('start')

        cmd_check_ping = 'grep -ire ad_cldap_ping_send ' \
                         '/var/log/sssd/sssd_%s.log | ' \
                         'grep -ire \"Found 2 domain controllers in domain ' \
                         'Default-First-Site-Name._sites.%s\"'\
                         % (domain, domain)
        check_ping = multihost.client[0].run_command(cmd_check_ping,
                                                     raiseonerr=False)
        assert check_ping.returncode == 0
        cmd_check_batch1 = 'grep -ire ad_cldap_ping_parallel_batch' \
                           ' /var/log/sssd/sssd_%s.log | ' \
                           'grep -ire \" %s\"' % (domain, ad1)
        check_batch1 = multihost.client[0].run_command(cmd_check_batch1,
                                                       raiseonerr=False)
        cmd_check_batch2 = 'grep -ire ad_cldap_ping_parallel_batch' \
                           ' /var/log/sssd/sssd_%s.log | ' \
                           'grep -ire \" %s\"' % (domain, ad2)
        check_batch2 = multihost.client[0].run_command(cmd_check_batch2,
                                                       raiseonerr=False)
        if check_batch1.returncode == 1 and check_batch2.returncode == 0:
            assert True
        else:
            assert False
        cmd_check_discovery = 'grep -ire ad_cldap_ping_domain_discovery_done' \
                              ' /var/log/sssd/sssd_%s.log | ' \
                              'grep -ire \"Found 2 domain' \
                              ' controllers in domain ' \
                              'Default-First-Site-Name._sites.%s\"'\
                              % (domain, domain)
        check_discovery = multihost.client[0].run_command(cmd_check_discovery,
                                                          raiseonerr=False)
        assert check_discovery.returncode == 0

        fw_stop = 'systemctl stop firewalld'
        multihost.client[0].run_command(fw_stop, raiseonerr=True)
        fw_remove = 'dnf remove -y firewalld'
        multihost.client[0].run_command(fw_remove, raiseonerr=True)

    @pytest.mark.adsites
    @pytest.mark.c_ares
    def test_003_ad_startup_discovery_two_different_sites(self, multihost,
                                                          adjoin, create_site):
        """
         @Title: IDM-SSSD-TC: ad_startup_discovery_two_different_sites
        * grep sssd domain logs for cldap ping
        * grep sssd logs for cldap ping parallel batch
        * grep sssd logs for cldap ping domain discovery
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = client.get_domain_section_name()
        domain_section = 'domain/{}'.format(domain)
        sssd_params = {'debug_level': '0xFFF0'}
        client.sssd_conf(domain_section, sssd_params)

        ad1 = multihost.ad[0].hostname
        ad2 = multihost.ad[1].hostname
        multihost.client[0].service_sssd('start')

        cmd_check_ping = 'grep -ire ad_cldap_ping_send' \
                         ' /var/log/sssd/sssd_%s.log | ' \
                         'grep -ire \"Found 2 domain controllers in domain ' \
                         'Default-First-Site-Name._sites.%s\"'\
                         % (domain, domain)
        check_ping = multihost.client[0].run_command(cmd_check_ping,
                                                     raiseonerr=False)
        assert check_ping.returncode == 0
        cmd_check_batch1 = 'grep -ire ad_cldap_ping_parallel_batch' \
                           ' /var/log/sssd/sssd_%s.log | ' \
                           'grep -ire \" %s\"' % (domain, ad1)
        check_batch1 = multihost.client[0].run_command(cmd_check_batch1,
                                                       raiseonerr=False)
        cmd_check_batch2 = 'grep -ire ad_cldap_ping_parallel_batch' \
                           ' /var/log/sssd/sssd_%s.log | ' \
                           'grep -ire \" %s\"' % (domain, ad2)
        check_batch2 = multihost.client[0].run_command(cmd_check_batch2,
                                                       raiseonerr=False)
        if check_batch1.returncode == 0 or check_batch2.returncode == 0:
            assert True
        else:
            assert False
        cmd_check_discovery = 'grep -ire ad_cldap_ping_domain_discovery_done' \
                              ' /var/log/sssd/sssd_%s.log | ' \
                              'grep -ire \"Found 2 domain' \
                              ' controllers in domain ' \
                              'Default-First-Site-Name._sites.%s\"'\
                              % (domain, domain)
        check_discovery = multihost.client[0].run_command(cmd_check_discovery,
                                                          raiseonerr=False)
        assert check_discovery.returncode == 0

    @pytest.mark.adsites
    @pytest.mark.c_ares
    def test_004_ad_startup_discovery_one_server_unreachable(self,
                                                             multihost,
                                                             adjoin,
                                                             create_site):
        """
        @Title: IDM-SSSD-TC:
        ad_startup_discovery_two_different_sites_one_server_unreachable
        * grep sssd domain logs for cldap ping
        * grep sssd logs for cldap ping parallel batch
        * grep sssd logs for cldap ping domain discovery
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = client.get_domain_section_name()
        domain_section = 'domain/{}'.format(domain)
        sssd_params = {'debug_level': '0xFFF0'}
        client.sssd_conf(domain_section, sssd_params)

        ad1 = multihost.ad[0].hostname
        ad2 = multihost.ad[1].hostname
        ad2ip = multihost.ad[1].ip

        cmd_dnf_firewalld = 'dnf install -y firewalld'
        multihost.client[0].run_command(cmd_dnf_firewalld)
        cmd_start_firewalld = 'systemctl start firewalld'
        multihost.client[0].run_command(cmd_start_firewalld)
        fw_add = 'firewall-cmd --permanent --direct --add-rule ipv4 ' \
                 'filter OUTPUT 0 -d %s -j DROP' % ad2ip
        fw_reload = 'firewall-cmd --reload'
        multihost.client[0].run_command(fw_add, raiseonerr=True)
        multihost.client[0].run_command(fw_reload, raiseonerr=True)

        multihost.client[0].service_sssd('start')

        cmd_check_ping = 'grep -ire ad_cldap_ping_send' \
                         ' /var/log/sssd/sssd_%s.log | ' \
                         'grep -ire \"Found 2 domain controllers in domain ' \
                         'Default-First-Site-Name._sites.%s\"'\
                         % (domain, domain)
        check_ping = multihost.client[0].run_command(cmd_check_ping,
                                                     raiseonerr=False)
        assert check_ping.returncode == 0
        cmd_check_batch1 = 'grep -ire ad_cldap_ping_parallel_batch' \
                           ' /var/log/sssd/sssd_%s.log | ' \
                           'grep -ire \" %s\"' % (domain, ad1)
        check_batch1 = multihost.client[0].run_command(cmd_check_batch1,
                                                       raiseonerr=False)
        cmd_check_batch2 = 'grep -ire ad_cldap_ping_parallel_batch' \
                           ' /var/log/sssd/sssd_%s.log | ' \
                           'grep -ire \" %s\"' % (domain, ad2)
        check_batch2 = multihost.client[0].run_command(cmd_check_batch2,
                                                       raiseonerr=False)
        if check_batch1.returncode == 1 and check_batch2.returncode == 0:
            assert True
        else:
            assert False
        cmd_check_discovery = 'grep -ire ad_cldap_ping_domain_discovery_done' \
                              ' /var/log/sssd/sssd_%s.log | ' \
                              'grep -ire \"Found 2 domain' \
                              ' controllers in domain ' \
                              'Default-First-Site-Name._sites.%s\"'\
                              % (domain, domain)
        check_discovery = multihost.client[0].run_command(cmd_check_discovery,
                                                          raiseonerr=False)
        assert check_discovery.returncode == 0

        fw_stop = 'systemctl stop firewalld'
        multihost.client[0].run_command(fw_stop, raiseonerr=True)
        fw_remove = 'dnf remove -y firewalld'
        multihost.client[0].run_command(fw_remove, raiseonerr=True)



import socket
import ssh2
from ssh2.session import Session


class SSHClient:
    """ ssh2 methods """
    def __init__(self, hostname, username, password):
        """Initialize defaults"""
        self.hostname = hostname
        self.username = username
        self.password = password
        self.session = None

    def connect(self):
        """login to host"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.hostname, 22))
        session = Session()
        session.handshake(sock)
        session.userauth_password(self.username, self.password)
        self.session = session

    def execute_command(self, command):
        """Run Non interactive Commands"""
        channel = self.session.open_session()
        channel.execute(command)
        size, data = channel.read()
        output = ""
        while size > 0:
            output += data.decode()
            size, data = channel.read()
        channel.close()
        return output

    def close(self):
        """Logout of ssh session"""
        if self.session:
            self.session.disconnect()


def check_login(hostname, user, password):
    """This function will check user login
    user: Name of the user.
    hostname: Name of the machine where user will login.
    password: User password.
    """
    ssh = SSHClient(hostname, user, password)
    ssh.connect()
    ssh.close()


def check_login_client(multihost, user, password):
    """This function will check user login
    user: Name of the user.
    password: User password.
    """
    hostname = multihost.client[0].ip
    ssh = SSHClient(hostname, user, password)
    ssh.connect()
    ssh.close()


def check_login_client_bool(multihost, user, password):
    """This function will check user login
    user: Name of the user.
    password: User password.
    """
    try:
        hostname = multihost.client[0].ip
        ssh = SSHClient(hostname, user, password)
        ssh.connect()
        ssh.close()
        return True
    except ssh2.exceptions.AuthenticationError:
        return False


def run_command(hostname, user, password, command):
    """This function will execute command
    user: Name of the user.
    hostname: Name of the machine where user will login.
    password: User password.
    command: User command
    """
    ssh = SSHClient(hostname, user, password)
    ssh.connect()
    result = ssh.execute_command(command)
    ssh.close()
    return result


def run_command_client(multihost, user, password, command):
    """This function will execute command
    user: Name of the user.
    hostname: Name of the machine where user will login.
    password: User password.
    command: User command
    """
    hostname = multihost.client[0].ip
    ssh = SSHClient(hostname, user, password)
    ssh.connect()
    result = ssh.execute_command(command)
    ssh.close()
    return result



#!/usr/bin/env python
#  SSSD
#
#  SSSD HBAC python API tests
#
#  Copyright (C) Red Hat
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import unittest
import sys
import os
import copy
import tempfile

BUILD_DIR = os.getenv('builddir') or "."
TEST_DIR = os.path.realpath(os.getenv('SSS_TEST_DIR') or ".")
MODPATH = tempfile.mkdtemp(prefix="tp_pyhbac_", dir=TEST_DIR)


if sys.version_info[0] > 2:
    unicode = str


def compat_assertIsInstance(this, obj, cls, msg=None):
    return this.assertTrue(isinstance(obj, cls))


def compat_assertItemsEqual(this, expected_seq, actual_seq, msg=None):
    return this.assertEqual(sorted(expected_seq), sorted(actual_seq))


# add compat assertIsInstance for old unittest.TestCase versions
# (python < 2.7, RHEL6 for instance)
if not hasattr(unittest.TestCase, "assertIsInstance"):
    setattr(unittest.TestCase, "assertIsInstance", compat_assertIsInstance)

# Python3 renamed assertItemsEqual to assertCountEqual but at the same time
# Python2 doesn't have assertCountEqual, see http://bugs.python.org/issue17866
if not hasattr(unittest.TestCase, "assertCountEqual"):
    if not hasattr(unittest.TestCase, "assertItemsEqual"):
        # This is RHEL-6
        setattr(unittest.TestCase, "assertItemsEqual", compat_assertItemsEqual)

    setattr(unittest.TestCase,
            "assertCountEqual",
            unittest.TestCase.assertItemsEqual)


class PyHbacImport(unittest.TestCase):
    def setUp(self):
        " Make sure we load the in-tree module "
        self.system_path = sys.path[:]
        sys.path = [MODPATH]

    def tearDown(self):
        " Restore the system path "
        sys.path = self.system_path

    def testImport(self):
        " Import the module and assert it comes from tree "
        try:
            dest_module_path = MODPATH + "/pyhbac.so"

            if sys.version_info[0] > 2:
                src_module_path = BUILD_DIR + "/.libs/_py3hbac.so"
            else:
                src_module_path = BUILD_DIR + "/.libs/_py2hbac.so"

            src_module_path = os.path.abspath(src_module_path)
            os.symlink(src_module_path, dest_module_path)

            import pyhbac
        except ImportError as e:
            print("Could not load the pyhbac module. Please check if it is "
                  "compiled", file=sys.stderr)
            raise e
        self.assertEqual(os.path.realpath(pyhbac.__file__),
                         os.path.realpath(MODPATH + "/pyhbac.so"))


class PyHbacRuleElementTest(unittest.TestCase):
    def testInstantiateEmpty(self):
        el = pyhbac.HbacRuleElement()
        self.assertCountEqual(el.names, [])
        self.assertCountEqual(el.groups, [])
        self.assertCountEqual(el.category, set([pyhbac.HBAC_CATEGORY_NULL]))

    def testInit(self):
        names = ["foo", "bar"]
        el = pyhbac.HbacRuleElement(names=names)
        self.assertCountEqual(el.names, names)

        groups = ["abc", "def"]
        el = pyhbac.HbacRuleElement(groups=groups)
        self.assertCountEqual(el.groups, groups)

    def testGetSet(self):
        names = ["foo", "bar"]
        el = pyhbac.HbacRuleElement()
        self.assertCountEqual(el.names, [])
        el.names = names
        self.assertCountEqual(el.names, names)

        groups = ["abc", "def"]
        el = pyhbac.HbacRuleElement()
        self.assertCountEqual(el.groups, [])
        el.groups = groups
        self.assertCountEqual(el.groups, groups)

        # Test other iterables than list
        groups = ("abc", "def")
        el = pyhbac.HbacRuleElement()
        self.assertCountEqual(el.groups, [])
        el.groups = groups
        self.assertCountEqual(el.groups, groups)

    def testCategory(self):
        el = pyhbac.HbacRuleElement()
        assert pyhbac.HBAC_CATEGORY_NULL in el.category
        assert pyhbac.HBAC_CATEGORY_ALL not in el.category

        el.category.add(pyhbac.HBAC_CATEGORY_ALL)
        assert pyhbac.HBAC_CATEGORY_ALL in el.category

        el.category = set([pyhbac.HBAC_CATEGORY_ALL])
        assert pyhbac.HBAC_CATEGORY_ALL in el.category

        # negative tests
        self.assertRaises(TypeError, el.__setattr__, "category",
                          [pyhbac.HBAC_CATEGORY_ALL])
        self.assertRaises(TypeError, el.__setattr__, "category", None)
        self.assertRaises(TypeError, el.__setattr__, "category", 1)

    def testNotIterable(self):
        self.assertRaises(TypeError, pyhbac.HbacRuleElement, names=123)
        self.assertRaises(TypeError, pyhbac.HbacRuleElement, names=None)

    def testRuleElementReference(self):
        def _get_rule():
            users = ["foo", "bar"]
            user_groups = ["abc", "def"]
            return pyhbac.HbacRuleElement(names=users, groups=user_groups)

        el = _get_rule()
        self.assertCountEqual(el.names, ["foo", "bar"])
        self.assertCountEqual(el.groups, ["abc", "def"])

    def testRepr(self):
        el = pyhbac.HbacRuleElement()
        self.assertEqual(el.__repr__(), u'<category 0 names [] groups []>')

        el.category.add(pyhbac.HBAC_CATEGORY_ALL)
        el.names = ['foo']
        el.groups = ['bar, baz']
        self.assertEqual(el.__repr__(),
                         u'<category 1 names [foo] groups [bar, baz]>')


class PyHbacRuleTest(unittest.TestCase):
    def testRuleGetSetName(self):
        name = "testGetRule"
        new_name = "testGetNewRule"

        rule = pyhbac.HbacRule(name)
        self.assertEqual(rule.name, unicode(name))

        rule.name = new_name
        self.assertEqual(rule.name, unicode(new_name))

    def testRuleGetSetEnabled(self):
        rule = pyhbac.HbacRule("testRuleGetSetEnabled")

        rule.enabled = True
        self.assertEqual(rule.enabled, True)
        rule.enabled = False
        self.assertEqual(rule.enabled, False)

        rule.enabled = "TRUE"
        self.assertEqual(rule.enabled, True)
        rule.enabled = "FALSE"
        self.assertEqual(rule.enabled, False)

        rule.enabled = "true"
        self.assertEqual(rule.enabled, True)
        rule.enabled = "false"
        self.assertEqual(rule.enabled, False)

        rule.enabled = "True"
        self.assertEqual(rule.enabled, True)
        rule.enabled = "False"
        self.assertEqual(rule.enabled, False)

        rule.enabled = 1
        self.assertEqual(rule.enabled, True)
        rule.enabled = 0
        self.assertEqual(rule.enabled, False)

        # negative test
        self.assertRaises(TypeError, rule.__setattr__, "enabled", None)
        self.assertRaises(TypeError, rule.__setattr__, "enabled", [])
        self.assertRaises(ValueError, rule.__setattr__, "enabled", "foo")
        self.assertRaises(ValueError, rule.__setattr__, "enabled", 5)

    def testRuleElementInRule(self):
        users = ["foo", "bar"]
        user_groups = ["abc", "def"]

        # rule should contain empty elements after instantiation
        rule = pyhbac.HbacRule("testRuleElement")
        self.assertIsInstance(rule.users, pyhbac.HbacRuleElement)
        self.assertIsInstance(rule.services, pyhbac.HbacRuleElement)
        self.assertIsInstance(rule.targethosts, pyhbac.HbacRuleElement)
        self.assertIsInstance(rule.srchosts, pyhbac.HbacRuleElement)

        self.assertIsInstance(rule.users.names, list)
        self.assertIsInstance(rule.users.groups, list)
        self.assertCountEqual(rule.users.names, [])
        self.assertCountEqual(rule.users.groups, [])

        # Assign by copying a HbacRuleElement
        user_el = pyhbac.HbacRuleElement(names=users, groups=user_groups)
        rule = pyhbac.HbacRule("testRuleElement")
        rule.users = user_el
        self.assertCountEqual(rule.users.names, users)
        self.assertCountEqual(rule.users.groups, user_groups)

        # Assign directly
        rule = pyhbac.HbacRule("testRuleElement")
        rule.users.names = users
        rule.users.groups = user_groups
        self.assertCountEqual(rule.users.names, users)
        self.assertCountEqual(rule.users.groups, user_groups)

    def testRuleElementInRuleReference(self):
        " Test that references to RuleElement are kept even if element goes"
        " out of scope "
        def _get_rule():
            users = ["foo", "bar"]
            user_groups = ["abc", "def"]
            el = pyhbac.HbacRuleElement(names=users, groups=user_groups)
            rule = pyhbac.HbacRule("testRuleElement")
            rule.users = el
            return rule

        rule = _get_rule()
        self.assertCountEqual(rule.users.names, ["foo", "bar"])
        self.assertCountEqual(rule.users.groups, ["abc", "def"])

    def testRepr(self):
        r = pyhbac.HbacRule('foo')
        self.assertEqual(r.__repr__(),
                         u"<name foo enabled 0 "
                         "users <category 0 names [] groups []> "
                         "services <category 0 names [] groups []> "
                         "targethosts <category 0 names [] groups []> "
                         "srchosts <category 0 names [] groups []>>")

        name = "someuser"
        service = "ssh"
        srchost = "host1"
        targethost = "host2"

        r.users.names = [name]
        r.services.names = [service]
        r.srchosts.names = [srchost]
        r.targethosts.names = [targethost]

        self.assertEqual(r.__repr__(),
                         u"<name foo enabled 0 "
                         "users <category 0 names [%s] groups []> "
                         "services <category 0 names [%s] groups []> "
                         "targethosts <category 0 names [%s] groups []> "
                         "srchosts <category 0 names [%s] groups []>>" %
                         (name, service, targethost, srchost))

    def testValidate(self):
        r = pyhbac.HbacRule('valid_rule')

        valid, missing = r.validate()
        self.assertEqual(valid, False)
        self.assertCountEqual(missing, (pyhbac.HBAC_RULE_ELEMENT_USERS,
                                        pyhbac.HBAC_RULE_ELEMENT_SERVICES,
                                        pyhbac.HBAC_RULE_ELEMENT_TARGETHOSTS,
                                        pyhbac.HBAC_RULE_ELEMENT_SOURCEHOSTS))

        r.users.names = ["someuser"]
        r.services.names = ["ssh"]

        valid, missing = r.validate()
        self.assertEqual(valid, False)
        self.assertCountEqual(missing, (pyhbac.HBAC_RULE_ELEMENT_TARGETHOSTS,
                                        pyhbac.HBAC_RULE_ELEMENT_SOURCEHOSTS))

        r.srchosts.names = ["host1"]
        r.targethosts.names = ["host2"]

        valid, missing = r.validate()
        self.assertEqual(valid, True)


class PyHbacRequestElementTest(unittest.TestCase):
    def testInstantiateEmpty(self):
        el = pyhbac.HbacRequestElement()
        self.assertCountEqual(el.name, "")
        self.assertCountEqual(el.groups, [])

    def testInit(self):
        name = "foo"
        el = pyhbac.HbacRequestElement(name=name)
        self.assertCountEqual(el.name, name)

        groups = ["abc", "def"]
        el = pyhbac.HbacRequestElement(groups=groups)
        self.assertCountEqual(el.groups, groups)

    def testGetSet(self):
        name = "foo"
        el = pyhbac.HbacRequestElement()
        self.assertCountEqual(el.name, "")
        el.name = name
        self.assertCountEqual(el.name, name)

        groups = ["abc", "def"]
        el = pyhbac.HbacRequestElement()
        self.assertCountEqual(el.groups, [])
        el.groups = groups
        self.assertCountEqual(el.groups, groups)

        # Test other iterables than list
        groups = ("abc", "def")
        el = pyhbac.HbacRequestElement()
        self.assertCountEqual(el.groups, [])
        el.groups = groups
        self.assertCountEqual(el.groups, groups)

    def testGroupsNotIterable(self):
        self.assertRaises(TypeError, pyhbac.HbacRequestElement, groups=None)
        self.assertRaises(TypeError, pyhbac.HbacRequestElement, groups=123)

    def testRepr(self):
        r = pyhbac.HbacRequestElement()
        self.assertEqual(r.__repr__(), u"<name  groups []>")

        r.name = 'foo'
        r.groups = ['bar', 'baz']
        self.assertEqual(r.__repr__(), u"<name foo groups [bar,baz]>")


class PyHbacRequestTest(unittest.TestCase):
    def testRequestElementHandling(self):
        name = "req_name"
        groups = ["g1", "g2"]

        # The request should be empty after instantiation
        req = pyhbac.HbacRequest()
        self.assertIsInstance(req.user, pyhbac.HbacRequestElement)
        self.assertIsInstance(req.service, pyhbac.HbacRequestElement)
        self.assertIsInstance(req.targethost, pyhbac.HbacRequestElement)
        self.assertIsInstance(req.srchost, pyhbac.HbacRequestElement)

        self.assertEqual(req.user.name, "")
        self.assertIsInstance(req.user.groups, list)
        self.assertCountEqual(req.user.groups, [])

        # Assign by copying a HbacRequestElement
        user_el = pyhbac.HbacRequestElement(name=name, groups=groups)
        req = pyhbac.HbacRequest()
        req.user = user_el
        self.assertCountEqual(req.user.name, name)
        self.assertCountEqual(req.user.groups, groups)

        # Assign directly
        req = pyhbac.HbacRequest()
        req.user.name = name
        req.user.groups = groups
        self.assertCountEqual(req.user.name, name)
        self.assertCountEqual(req.user.groups, groups)

    def testRuleName(self):
        req = pyhbac.HbacRequest()
        self.assertEqual(req.rule_name, None)
        # python 2.4 raises TypError, 2.7 raises AttributeError
        self.assertRaises((TypeError, AttributeError), req.__setattr__,
                          "rule_name", "foo")

    def testEvaluate(self):
        name = "someuser"
        service = "ssh"
        srchost = "host1"
        targethost = "host2"

        allow_rule = pyhbac.HbacRule("allowRule", enabled=True)
        allow_rule.users.names = [name]
        allow_rule.services.names = [service]
        allow_rule.srchosts.names = [srchost]
        allow_rule.targethosts.names = [targethost]

        req = pyhbac.HbacRequest()
        req.user.name = name
        req.service.name = service
        req.srchost.name = srchost
        req.targethost.name = targethost

        # Test that an allow rule on its own allows access
        res = req.evaluate((allow_rule,))
        self.assertEqual(res, pyhbac.HBAC_EVAL_ALLOW)
        self.assertEqual(req.rule_name, "allowRule")

        # Test that a user not in the rule is not allowed
        req.user.name = "someotheruser"
        res = req.evaluate((allow_rule, ))
        self.assertEqual(res, pyhbac.HBAC_EVAL_DENY)
        self.assertEqual(req.rule_name, None)

        # But allows if the rule is an ALL rule
        allow_rule.users.category.add(pyhbac.HBAC_CATEGORY_ALL)
        res = req.evaluate((allow_rule, ))
        self.assertEqual(res, pyhbac.HBAC_EVAL_ALLOW)

    def testRepr(self):
        name = "someuser"
        service = "ssh"
        srchost = "host1"
        targethost = "host2"

        req = pyhbac.HbacRequest()

        self.assertEqual(req.__repr__(), "<user <name  groups []> "
                                         "service <name  groups []> "
                                         "targethost <name  groups []> "
                                         "srchost <name  groups []>>")

        req.user.name = name
        req.service.name = service
        req.srchost.name = srchost
        req.targethost.name = targethost

        self.assertEqual(req.__repr__(), "<user <name %s groups []> "
                                         "service <name %s groups []> "
                                         "targethost <name %s groups []> "
                                         "srchost <name %s groups []>>" %
                                         (name, service, targethost, srchost))

    def testEvaluateNegative(self):
        name = "someuser"
        service = "ssh"
        srchost = "host1"
        targethost = "host2"

        allow_rule = pyhbac.HbacRule("allowRule", enabled=True)
        allow_rule.users.names = [name]
        allow_rule.services.names = [service]
        allow_rule.srchosts.names = [srchost]
        allow_rule.targethosts.names = [targethost]

        req = pyhbac.HbacRequest()
        req.service.name = service
        req.srchost.name = srchost
        req.targethost.name = targethost
        req.user.name = name

        saveuser = req.user
        req.user = None  # need to catch this

        # catch invalid category value
        savecat = copy.copy(allow_rule.users.category)
        allow_rule.users.category.add(pyhbac.HBAC_EVAL_ERROR)
        self.assertRaises(ValueError, req.evaluate, (allow_rule,))
        allow_rule.users.category = savecat

        # Test that invalid type is raised
        self.assertRaises(TypeError, req.evaluate, (allow_rule,))

        req.user = saveuser
        allow_rule.users = None  # need to catch this
        self.assertRaises(TypeError, req.evaluate, (allow_rule,))

        # catch invalid rule type
        self.assertRaises(TypeError, req.evaluate, (allow_rule, None))


class PyHbacModuleTest(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        os.unlink(MODPATH + "/pyhbac.so")
        os.rmdir(MODPATH)

    def testHasResultTypes(self):
        assert hasattr(pyhbac, "HBAC_EVAL_ALLOW")
        assert hasattr(pyhbac, "HBAC_EVAL_DENY")
        assert hasattr(pyhbac, "HBAC_EVAL_ERROR")

    def testHasErrorTypes(self):
        assert hasattr(pyhbac, "HBAC_ERROR_UNKNOWN")
        assert hasattr(pyhbac, "HBAC_SUCCESS")
        assert hasattr(pyhbac, "HBAC_ERROR_NOT_IMPLEMENTED")
        assert hasattr(pyhbac, "HBAC_ERROR_OUT_OF_MEMORY")
        assert hasattr(pyhbac, "HBAC_ERROR_UNPARSEABLE_RULE")

    def testHasCategories(self):
        assert hasattr(pyhbac, "HBAC_CATEGORY_NULL")
        assert hasattr(pyhbac, "HBAC_CATEGORY_ALL")

    def testHasRuleElementTypes(self):
        assert hasattr(pyhbac, "HBAC_RULE_ELEMENT_USERS")
        assert hasattr(pyhbac, "HBAC_RULE_ELEMENT_SERVICES")
        assert hasattr(pyhbac, "HBAC_RULE_ELEMENT_TARGETHOSTS")
        assert hasattr(pyhbac, "HBAC_RULE_ELEMENT_SOURCEHOSTS")

    def testHbacResultString(self):
        results = [pyhbac.HBAC_EVAL_ALLOW, pyhbac.HBAC_EVAL_DENY,
                   pyhbac.HBAC_EVAL_ERROR]
        for r in results:
            s = pyhbac.hbac_result_string(r)
            self.assertIsInstance(s, unicode)
            assert len(s) > 0

    def testHbacErrorString(self):
        errors = [pyhbac.HBAC_ERROR_UNKNOWN,
                  pyhbac.HBAC_SUCCESS,
                  pyhbac.HBAC_ERROR_NOT_IMPLEMENTED,
                  pyhbac.HBAC_ERROR_OUT_OF_MEMORY,
                  pyhbac.HBAC_ERROR_UNPARSEABLE_RULE]
        for e in errors:
            s = pyhbac.hbac_error_string(e)
            self.assertIsInstance(s, unicode)
            assert len(s) > 0


if __name__ == "__main__":
    error = 0

    suite = unittest.TestLoader().loadTestsFromTestCase(PyHbacImport)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x1
        # need to bail out here because pyhbac could not be imported
        sys.exit(error)

    # import the pyhbac module into the global namespace, but make sure it's
    # the one in tree
    sys.path.insert(0, MODPATH)
    import pyhbac

    loadTestsFromTestCase = unittest.TestLoader().loadTestsFromTestCase

    suite = loadTestsFromTestCase(PyHbacRuleElementTest)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x2

    suite = loadTestsFromTestCase(PyHbacRuleTest)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x3

    suite = loadTestsFromTestCase(PyHbacRequestElementTest)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x4

    suite = loadTestsFromTestCase(PyHbacRequestTest)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x5

    suite = loadTestsFromTestCase(PyHbacModuleTest)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x6

    sys.exit(error)



"""
    Provide Exceptions for py.test framework
"""


class StandardException(Exception):
    """ Overrides Exception class """

    def __init__(self, msg=None, rval=1):
        if msg is None:
            msg = 'Error'
        self.msg = msg
        self.rval = rval
        super(StandardException, self).__init__(self.msg)

    def __str__(self):
        return "{} ({})".format(self.msg, self.rval)


class InvalidInput(StandardException):
    """
    Override StandardException used mainly when invalid input is passed
    """


class DirSrvException(StandardException):
    """
    Override StandardException, This exception is to be used for
    Directory Server related Errors
    """


class PkiLibException(StandardException):
    """
    Override StandardException,
    This exception is to be used for PKI/SSL related Errors
    """


class OSException(StandardException):
    """
    Override StandardException, This exception is to be used for
    Operating system errors.
    """


class LdapException(StandardException):
    """
    Override StandardException, This exception is to be used for LDAP Errors
    """


class RPMException(StandardException):
    """
    Override StandardException, This exception is to be used for RPM Errors
    """


class SSSDException(StandardException):
    """
     Override StandardException, This exception is to be used for SSSD Errors
    """


class SSHLoginException(StandardException):
    """
    Override StandardException,
    This exception is to be used for SSH Login Errors
    """



""" Automation for RFE related to --add-samba-data to adcli

:casecomponent: sssd
:requirement: Add support for passing --add-samba-data to adcli
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import subprocess
import time
import pytest
from sssd.testlib.common.utils import sssdTools


@pytest.mark.usefixtures('winbind_server', 'configure_samba')
@pytest.mark.smbsecretrotation
class Testsmbsecretrotation(object):

    @pytest.mark.tier2
    def test_0001_rotation(self, multihost):
        """
        :title: Verify machine passwd updates local smb secrets
        :id: 3d08ea1c-6724-4bc9-ac62-b8be66486ee4
        """
        # Get current hash of /var/lib/samba/private/secrets.tdb
        hash_cmd = 'sha1hmac /var/lib/samba/private/secrets.tdb'
        cmd = multihost.client[0].run_command(hash_cmd, raiseonerr=False)
        before_hash = cmd.stdout_text.strip()
        print("hash before reseting machine passwd", before_hash)
        stat = 'stat /var/lib/samba/private/secrets.tdb'
        cmd = multihost.client[0].run_command(stat, raiseonerr=False)
        # Get tdb-dump
        tdbdump = 'tdbdump /var/lib/samba/private/secrets.tdb'
        multihost.client[0].run_command(tdbdump, raiseonerr=False)
        client = sssdTools(multihost.client[0], multihost.ad[0])
        sssd_params = {'ad_maximum_machine_account_password_age': '1',
                       'ad_machine_account_password_renewal_opts': '300:15',
                       'ad_update_samba_machine_account_password': 'True',
                       'debug_level': '9'}
        domain_name = client.get_domain_section_name()
        domain_section = 'domain/{}'.format(domain_name)
        client.sssd_conf(domain_section, sssd_params)
        client.clear_sssd_cache()
        client.reset_machine_password()
        client_hostname = multihost.client[0].sys_hostname.split('.')[0]
        if len(client_hostname) > 15:
            client_hostname = client_hostname[:15]
        realm = multihost.ad[0].realm
        host_princ = 'HOST/%s@%s' % (client_hostname, realm)
        kvno_cmd = 'kvno %s' % (host_princ)
        cmd = multihost.client[0].run_command(kvno_cmd, raiseonerr=False)
        kvno = cmd.stdout_text.split('=')[1].strip()
        restart_sssd = 'systemctl restart sssd'
        try:
            multihost.client[0].run_command(restart_sssd)
        except subprocess.CalledProcessError:
            multihost.client[0].multihost.client[0].run_command(
                'journalctl -x -n 50 --no-pager -u sssd', raiseonerr=False)
            pytest.fail("Cannot restart sssd service")
        time.sleep(30)
        ls = 'cat /etc/sssd/sssd.conf'
        cmd = multihost.client[0].run_command(ls, raiseonerr=False)
        klist_cmd = "klist -k /etc/krb5.keytab"
        cmd = multihost.client[0].run_command(klist_cmd, raiseonerr=False)
        spn_list = [val.strip() for val in cmd.stdout_text.splitlines()]
        client_domain = multihost.client[0].sys_hostname.split('.')[0].upper()
        if len(client_domain) > 15:
            client_domain = client_domain[:15]
        old_kvno = int(kvno)
        new_kvno = int(kvno) + 1
        # older entry
        entry_1 = '{} {}/{}@{}'.format(old_kvno, 'host', client_domain,
                                       multihost.ad[0].realm)
        # new entry
        entry_2 = '{} {}/{}@{}'.format(new_kvno, 'host', client_domain,
                                       multihost.ad[0].realm)
        if entry_1 and entry_2 not in spn_list[3:]:
            pytest.fail("keytab rotation failed, host entries not rotated")
        cmd = multihost.client[0].run_command(hash_cmd, raiseonerr=False)
        after_hash = cmd.stdout_text.strip()
        print("hash after reseting machine passwd", after_hash)
        cmd = multihost.client[0].run_command(stat, raiseonerr=False)
        logs = "tail -n 500 /var/log/sssd/sssd_%s.log" % domain_name
        multihost.client[0].run_command(logs, raiseonerr=False)
        multihost.client[0].run_command(tdbdump, raiseonerr=False)



#
# Module for simulation of utility "getent hosts -s sss" from coreutils
#
#   Authors:
#       Samuel Cabrero <scabrero@suse.com>
#
#   Copyright (C) 2019 SUSE LINUX GmbH, Nuernberg, Germany.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ctypes import (c_int, c_char_p, c_ulong, POINTER,
                    Structure, create_string_buffer)
from sssd_nss import NssReturnCode, SssdNssError, nss_sss_ctypes_loader
import socket
from ipaddress import IPv4Address, IPv6Address

HOST_BUFLEN = 1024


class Hostent(Structure):
    _fields_ = [("h_name", c_char_p),
                ("h_aliases", POINTER(c_char_p)),
                ("h_addrtype", c_int),
                ("h_length", c_int),
                ("h_addr_list", POINTER(c_char_p))]


def gethostbyname_r(name, result_p, buffer_p, buflen):
    """
    ctypes wrapper for:
        enum nss_status _nss_sss_gethostbyname_r(const char *name,
                                                 struct hostent *result,
                                                 char *buffer,
                                                 size_t buflen,
                                                 int *errnop,
                                                 int *h_errnop)
    """
    func = nss_sss_ctypes_loader("_nss_sss_gethostbyname_r")
    func.restype = c_int
    func.argtypes = [c_char_p, POINTER(Hostent),
                     c_char_p, c_ulong, POINTER(c_int), POINTER(c_int)]

    errno = POINTER(c_int)(c_int(0))
    h_errno = POINTER(c_int)(c_int(0))

    name = name.encode('utf-8')
    res = func(c_char_p(name), result_p, buffer_p, buflen, errno, h_errno)

    return (int(res), int(errno[0]), int(h_errno[0]), result_p)


def gethostbyname2_r(name, af, result_p, buffer_p, buflen):
    """
    ctypes wrapper for:
        enum nss_status _nss_sss_gethostbyname2_r(const char *name,
                                                  int af,
                                                  struct hostent *result,
                                                  char *buffer,
                                                  size_t buflen,
                                                  int *errnop,
                                                  int *h_errnop)
    """
    func = nss_sss_ctypes_loader("_nss_sss_gethostbyname2_r")
    func.restype = c_int
    func.argtypes = [c_char_p, c_int, POINTER(Hostent),
                     c_char_p, c_ulong, POINTER(c_int), POINTER(c_int)]

    errno = POINTER(c_int)(c_int(0))
    h_errno = POINTER(c_int)(c_int(0))

    name = name.encode('utf-8')
    res = func(c_char_p(name), af, result_p, buffer_p, buflen, errno, h_errno)

    return (int(res), int(errno[0]), int(h_errno[0]), result_p)


def set_hostent_dict(res, result_p):
    if res != NssReturnCode.SUCCESS:
        return dict()

    hostent_dict = dict()
    hostent_dict['name'] = result_p[0].h_name.decode('utf-8')
    hostent_dict['aliases'] = list()
    hostent_dict['addrtype'] = result_p[0].h_addrtype
    hostent_dict['length'] = result_p[0].h_length
    hostent_dict['addresses'] = list()

    i = 0
    while result_p[0].h_aliases[i] is not None:
        alias = result_p[0].h_aliases[i].decode('utf-8')
        hostent_dict['aliases'].append(alias)
        i = i + 1

    i = 0
    while result_p[0].h_addr_list[i] is not None:
        length = result_p[0].h_length
        binaddr = result_p[0].h_addr_list[i][:length]
        if result_p[0].h_addrtype == socket.AF_INET:
            addr = IPv4Address(binaddr)
            addr = socket.inet_ntop(socket.AF_INET, addr.packed)
        elif result_p[0].h_addrtype == socket.AF_INET6:
            addr = IPv6Address(binaddr)
            addr = socket.inet_ntop(socket.AF_INET, addr.packed)
        else:
            raise Exception("Failed to parse IP address")

        hostent_dict['addresses'].append(addr)
        i = i + 1

    return hostent_dict


def call_sssd_gethostbyname(name):
    """
    A Python wrapper to retrieve a host by name. Returns:
        (res, hostent_dict)
    if res is NssReturnCode.SUCCESS, then hostent_dict contains the keys
    corresponding to the C hostent structure fields. Otherwise, the dictionary
    is empty and errno indicates the error code
    """
    result = Hostent()
    result_p = POINTER(Hostent)(result)
    buff = create_string_buffer(HOST_BUFLEN)

    (res, errno, h_errno, result_p) = gethostbyname_r(name, result_p,
                                                      buff, HOST_BUFLEN)
    if errno != 0:
        raise SssdNssError(errno, "gethostbyname_r")

    hostent_dict = set_hostent_dict(res, result_p)
    return (res, h_errno, hostent_dict)



"""
SSSD Failover tests.

:requirement: Failover
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.ad import AD
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericProvider
from sssd_test_framework.roles.ipa import IPA
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.roles.samba import Samba
from sssd_test_framework.topology import KnownTopologyGroup


@pytest.mark.parametrize("value, expected", [(None, 31), (15, 31), (60, 60)])
@pytest.mark.importance("low")
@pytest.mark.ticket(gh=7375, jira="RHEL-17659")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_failover__reactivation_timeout_is_honored(
    client: Client, provider: GenericProvider, value: int | None, expected: int
):
    """
    :title: Primary server reactivation timeout is honored
    :setup:
        1. Create user "user-1"
        2. Set failover_primary_timeout to @value
        3. Set server/URI to an invalid server
        4. Set backup server/URI to the server
        5. Start SSSD
    :steps:
        1. Lookup user-1
        2. Check that SSSD is connected to the backup server
        3. Find "Primary server reactivation timeout set to @expected seconds" in domain logs
    :expectedresults:
        1. User is found
        2. SSSD is connected to the backup server
        3. String is found
    :customerscenario: True
    """
    provider.user("user-1").add()

    if value is not None:
        client.sssd.domain["failover_primary_timeout"] = str(value)

    client.sssd.enable_responder("ifp")

    if isinstance(provider, LDAP):
        client.sssd.domain["ldap_uri"] = "ldap://ldap.invalid"
        client.sssd.domain["ldap_backup_uri"] = f"ldap://{provider.host.hostname}"

    if isinstance(provider, AD):
        client.sssd.domain["ad_server"] = "invalid.ad.test"
        client.sssd.domain["ad_backup_server"] = f"{provider.host.hostname}"

    if isinstance(provider, Samba):
        client.sssd.domain["ad_server"] = "invalid.samba.test"
        client.sssd.domain["ad_backup_server"] = f"{provider.host.hostname}"

    if isinstance(provider, IPA):
        client.sssd.domain["ipa_server"] = "invalid.ipa.test"
        client.sssd.domain["ipa_backup_server"] = f"{provider.host.hostname}"

    client.sssd.start()

    # Lookup user to make sure SSSD did correctly failover to the backup server
    result = client.tools.id("user-1")
    assert result is not None, "User is not found!"

    # Check that SSSD is indeed connected to the backup server
    assert client.sssd.default_domain is not None, "Default domain is not set!"
    status = client.sssctl.domain_status(client.sssd.default_domain, active=True)
    assert provider.host.hostname in status.stdout, f"{provider.host.hostname} is not found in domain status!"

    # Check that primary server reactivation timeout was correctly created
    log = client.fs.read(client.sssd.logs.domain())
    assert (
        f"Primary server reactivation timeout set to {expected} seconds" in log
    ), f"'Primary server reactivation timeout set to {expected} seconds' not found in logs!"



#
# Abstract directory server instance class
#
# Copyright (c) 2015 Red Hat, Inc.
# Author: Nikolai Kondrashov <Nikolai.Kondrashov@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ldap


class DS(object):
    """Abstract directory server instance."""

    def __init__(self, dir, port, base_dn, admin_rdn, admin_pw):
        """
            Initialize the instance.

            Arguments:
            dir         Path to the root of the filesystem hierarchy to create
                        the instance under.
            port        TCP port on localhost to bind the server to.
            base_dn     Base DN.
            admin_rdn   Administrator DN, relative to BASE_DN.
            admin_pw    Administrator password.
        """
        self.dir = dir
        self.port = port
        self.ldap_url = "ldap://localhost:" + str(self.port)
        self.base_dn = base_dn
        self.admin_rdn = admin_rdn
        self.admin_dn = admin_rdn + "," + base_dn
        self.admin_pw = admin_pw

    def setup(self):
        """Setup the instance"""
        raise NotImplementedError()

    def teardown(self):
        """Teardown the instance"""
        raise NotImplementedError()

    def bind(self):
        """Connect to the server and bind as admin, return connection."""
        conn = ldap.initialize(self.ldap_url)
        conn.simple_bind_s(self.admin_dn, self.admin_pw)
        return conn



from __future__ import print_function

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
import tempfile
import os
import subprocess


class krb5srv(object):
    """ Base class to setup MIT Kerberos server
        Default Kerberos Server realm is "EXAMPLE.TEST"
        Default Admin password is "Secret123"
    """

    def __init__(self, multihost, krb_realm=None):
        """ Initialize realm, host, ports, Kerberos admin password """
        if krb_realm is None:
            self.krb_realm = 'EXAMPLE.TEST'
        else:
            self.krb_realm = krb_realm.upper()
        self.multihost = multihost
        self.kdc_port = '88'
        self.kadmin_port = '749'
        self.admin_password = 'Secret123'
        self.krb5_kdc_data_dir = '/var/kerberos/krb5kdc'
        self.krb5_kdc_sysconfig = '/etc/sysconfig/krb5kdc'
        self.krb5_log_file = '/var/log/krb5kdc.log'
        self.admin_keytab = '%s/kadm5.keytab' % (self.krb5_kdc_data_dir)
        self.kadmin_log_file = '/var/log/kadmind.log'
        self.krb_acl_file = '%s/kadm5.acl' % (self.krb5_kdc_data_dir)
        self.admin_keytab = '%s/kadm5.keytab' % (self.krb5_kdc_data_dir)
        self.kdc_conf = '%s/kdc.conf' % (self.krb5_kdc_data_dir)
        self.ciphers = "aes256-cts-hmac-sha384-192:normal "\
            "aes128-cts-hmac-sha256-128:normal "\
            "aes256-cts-hmac-sha1-96:normal "\
            "aes128-cts-hmac-sha1-96:normal"

    def _config_krb5kdc(self):
        """ Configure kdc.conf and kadm5.acl
            :param: None
            :return str: Return Kerberos kdc.conf file path
        """

        realm_def = f""" {{
        acl_file = {self.krb_acl_file}
        admin_keytab = {self.admin_keytab}
        supported_enctypes = {self.ciphers}
        }}"""
        config = ConfigParser.RawConfigParser()
        config.optionxform = str
        config.add_section('kdcdefaults')
        config.set('kdcdefaults', 'kdc_ports', self.kdc_port)
        config.set('kdcdefaults', 'kdc_tcp_ports', self.kdc_port)
        config.add_section('logging')
        config.set('logging', 'kdc', 'FILE:%s' % (self.krb5_log_file))
        config.set('logging', 'admin_server', 'File:%s' %
                   (self.kadmin_log_file))
        config.add_section('realms')
        config.set('realms', self.krb_realm, realm_def)

        (krb_config, krb_config_path) = tempfile.mkstemp(suffix='cfg')
        os.close(krb_config)
        with open(krb_config_path, "w") as outfile:
            config.write(outfile)
        return krb_config_path

    def _krb_acl_config(self):
        """ Create ACL file
            :param: None
            :return: Config file path
        """
        acl = "*/%s@%s *" % ('admin', self.krb_realm)
        (acl_config, acl_config_path) = tempfile.mkstemp(suffix='cfg')
        os.close(acl_config)
        with open(acl_config_path, "w") as outfile:
            outfile.write(acl)
        return acl_config_path

    def krb_setup_new(self):
        """ Setup new Kerberos REALM
            :param: None
            :return: None
            :Exception: Raises subprocess.CalledProcessError
        """
        krb_config_path = self._config_krb5kdc()
        acl_file = self._krb_acl_config()
        self.multihost.transport.put_file(krb_config_path, self.kdc_conf)
        self.multihost.transport.put_file(acl_file, self.krb_acl_file)
        kdb5_util_create = ['kdb5_util', 'create', '-W', '-r',
                            self.krb_realm, '-s', '-P', '""']

        try:
            self.multihost.run_command(kdb5_util_create)
        except subprocess.CalledProcessError:
            raise

        try:
            self.add_principal(p_type='admin',
                               password=self.admin_password,
                               service='admin',
                               service_name='admin')
        except subprocess.CalledProcessError:
            raise
        else:
            self.multihost.log.info("created REALM %s" % (self.krb_realm))

        try:
            self.add_principal(p_type='admin',
                               password=self.admin_password,
                               service='root')
        except subprocess.CalledProcessError:
            raise

        try:
            self.add_principal(p_type=None, service='host',
                               service_name=self.multihost.sys_hostname)
        except subprocess.CalledProcessError:
            raise
        else:
            self.multihost.log.info("host principal added")

        try:
            self.multihost.run_command(['kadmin.local', '-r', self.krb_realm,
                                        '-q', "ktadd host/%s" %
                                        (self.multihost.sys_hostname)])
        except subprocess.CalledProcessError:
            raise
        try:
            self.multihost.run_command(['kadmin.local', '-r', self.krb_realm,
                                        '-q', "ktadd -k %s kadmin/admin" %
                                        (self.admin_keytab)])
        except subprocess.CalledProcessError:
            raise

        try:
            self.multihost.run_command(['kadmin.local', '-r', self.krb_realm,
                                        '-q', "ktadd -k %s kadmin/changepw" %
                                        (self.admin_keytab)])
        except subprocess.CalledProcessError:
            raise

        try:
            self.multihost.run_command(['systemctl', 'start', 'krb5kdc'])
        except subprocess.CalledProcessError:
            raise
        else:
            self.multihost.log.info("krb5kdc service start successful")
        try:
            self.multihost.run_command(['systemctl', 'start', 'kadmin'])
        except subprocess.CalledProcessError:
            raise
        else:
            self.multihost.log.info("kadmin service started successfully")

    def add_principal(self, principal=None,
                      p_type='user',
                      password=None,
                      service=None,
                      service_name=None,
                      etype=None):
        """ Add server/user principals to Kerberos server
            :param str principal: principal name (foobar)
            :param str p_type: principal type (user/admin/None)
            :param str password: password ('Secret123')
            :param str service: service principal (host/http/nfs)
            :param str service_name: Hostname where service is run
            :return bool: True if principal is added
            :Exception: Raise subprocess.CalledProcessError
        """
        # Todo: Need to check if a principal already exists before adding.
        if service is None:
            service = 'host'

        if p_type == 'user':
            add_principal = "add_principal -clearpolicy"\
                            " -pw %s %s@%s" % (password, principal,
                                               self.krb_realm)
            if etype:
                add_principal = "add_principal -clearpolicy"\
                                " -e %s -pw %s %s@%s" % (etype, password,
                                                         principal,
                                                         self.krb_realm)
            kadmin_local_cmd = ['kadmin.local', '-r',
                                self.krb_realm, '-q', add_principal]
        elif p_type == 'admin':
            add_principal = "add_principal -clearpolicy"\
                            " -pw %s %s/%s" % (password, service, 'admin')
            kadmin_local_cmd = ['kadmin.local', '-r', self.krb_realm,
                                '-q', add_principal]
        else:
            add_principal = "add_principal -clearpolicy"\
                            " -randkey %s/%s" % (service, service_name)
            kadmin_local_cmd = ['kadmin.local', '-r', self.krb_realm,
                                '-q', add_principal]
        try:
            self.multihost.run_command(kadmin_local_cmd)
        except subprocess.CalledProcessError:
            raise
        else:
            return True

    def delete_principal(self, principal):
        """ Delete kerberos principal
        :param str principal: principal name (foobar)
        :return bool: True if principal is deleted
        :Exception: Raise subprocess.CalledProcessError
        """
        del_principal = "delete_principal -force principal"
        kadmin_local_cmd = ['kadmin.local', '-r',
                            self.krb_realm, '-q', del_principal]
        try:
            self.multihost.run_command(kadmin_local_cmd)
        except subprocess.CalledProcessError:
            raise
        else:
            return True

    def add_remove_keytab(self, comp_principal=None,
                          location=None,
                          operation=None):
        """ Add or Remove keytab  to Kerberos principal
            :param str com_principal: complete comp_principal
             example(abc@EXAMPLE.COM)
            :param str location: location of keytab file with
             .keytab extension ('/etc/krb5.keytab)
            :param str operation: Value must add or remove
            :return bool: True if keytab is added
            :Exception: Raise subprocess.CalledProcessError
        """

        if operation == 'add':
            ops = 'ktadd'

        elif operation == 'remove':
            ops = 'ktremove'

        else:
            return False

        if location is None:
            location = '/etc/krb5.keytab'

        add_keytab = "%s -k "\
            "%s %s@%s" % (ops, location, comp_principal,
                          self.krb_realm)
        kadmin_local_cmd = ['kadmin.local', '-r',
                            self.krb_realm, '-q', add_keytab]

        try:
            self.multihost.run_command(kadmin_local_cmd)
        except subprocess.CalledProcessError:
            raise
        else:
            return True

    def destroy_krb5server(self):
        """ Destroy Kerberos database
            :param: None
            :Exception: subprocess.CalledProcessError
        """
        # stop the Kerberos server
        for service in ('krb5kdc', 'kadmin'):
            stop_cmd = 'systemctl stop %s' % service
            try:
                self.multihost.run_command(stop_cmd)
            except subprocess.CalledProcessError:
                raise
            else:
                self.multihost.log.info("stopped %s service ")

        # destroy Kerberos database
        try:
            self.multihost.run_command(['kdb5_util', 'destroy', '-f',
                                        self.krb_realm])
        except subprocess.CalledProcessError:
            raise
        else:
            self.multihost.log.info("Removed krb realm %s" % self.krb_realm)



"""
KCM responder tests.

:requirement: IDM-SSSD-REQ :: SSSD KCM as default Kerberos CCACHE provider
"""

from __future__ import annotations

import time

import pytest
from pytest_mh.conn import ProcessError
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.kdc import KDC
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.topology import KnownTopology


@pytest.mark.importance("critical")
@pytest.mark.authentication
@pytest.mark.topology(KnownTopology.Client)
@pytest.mark.parametrize("ccache_storage", ["memory", "secdb"])
def test_kcm__kinit_does_not_create_new_ccache(client: Client, kdc: KDC, ccache_storage: str):
    """
    :title: Second call to kinit with the same principal does not create new ccache.
    :setup:
        1. Add Kerberos principal "tuser" to KDC
        2. Add local user "tuser"
        3. Set ccache_storage in sssd.conf to @ccache_storage
        4. Start SSSD
    :steps:
        1. Authenticate as "tuser" over SSH
        2. Count existing credential caches
        3. Kinit as "tuser"
        4. Check that TGT was acquired
        5. Count existing credential caches
        6. Repeat steps 3-5
    :expectedresults:
        1. User is logged into the host
        2. Returns 0, no ccache is available
        3. User obtains TGT
        4. TGT is present
        5. Returns 1, single ccache is available
        6. Same results as for steps 3-5
    :customerscenario: False
    """
    kdc.principal("tuser").add(password="Secret123")
    client.local.user("tuser").add(password="Secret123")

    client.sssd.common.kcm(kdc)
    client.sssd.kcm["ccache_storage"] = ccache_storage
    client.sssd.start()

    with client.ssh("tuser", "Secret123") as ssh:
        with client.auth.kerberos(ssh) as krb:
            assert krb.cache_count() == 0, "KRB cache is not empty!"

            assert krb.kinit("tuser", password="Secret123").rc == 0, "kinit failed!"
            assert krb.has_tgt("tuser", kdc.realm), "No ticket found!"
            assert krb.cache_count() == 1, "KRB cache value is not 1!"

            assert krb.kinit("tuser", password="Secret123").rc == 0, "kinit failed!"
            assert krb.has_tgt("tuser", kdc.realm), "No ticket found!"
            assert krb.cache_count() == 1, "KRB cache value is not 1!"


@pytest.mark.importance("critical")
@pytest.mark.authentication
@pytest.mark.topology(KnownTopology.Client)
@pytest.mark.parametrize("ccache_storage", ["memory", "secdb"])
def test_kcm__ccache_holds_multiple_and_all_types_of_principals(client: Client, kdc: KDC, ccache_storage: str):
    """
    :title: Multiple principals and service tickets can be stored in a ccache collection.
    :setup:
        1. Add Kerberos principal "alice" to KDC
        2. Add Kerberos principal "bob" to KDC
        3. Add Kerberos principal "carol" to KDC
        4. Add Kerberos service principal "host/myhost" to KDC
        5. Add local user "tuser"
        6. Set ccache_storage in sssd.conf to @ccache_storage
        7. Start SSSD
    :steps:
        1. Authenticate as "tuser" over SSH
        2. Count existing credential caches
        3. Kinit as "alice"
        4. Kinit as "bob"
        5. Kinit as "carol"
        6. Obtain ticket for "host/myhost"
        7. Destroy current primary ccache
        8. Kinit as "carol"
        9. Kdestroy all ccaches
    :expectedresults:
        1. User is logged into the host
        2. Returns 0, no ccache is available
        3. 1 ccache exists, "alice" is the primary ccache, TGT is only ticket in "alice" ccache
        4. 2 ccaches exist, "bob" is the primary ccache, TGT is only ticket in "bob" ccache
        5. 3 ccaches exist, "carol" is the primary ccache, TGT is only ticket in "carol" ccache
        6. 3 ccaches exist, "carol" is the primary ccache, TGT is only ticket in "alice" and "bob" ccache,
           TGT and "host/myhost" are only tickets in "carol" ccache
        7. 2 ccaches exit
        8. 3 ccaches exist, "carol" is the primary ccache, TGT is only ticket in "alice", "bob" and "carol" ccache
        9. No ccache is available
    :customerscenario: False
    """
    kdc.principal("alice").add(password="Secret123")
    kdc.principal("bob").add(password="Secret123")
    kdc.principal("carol").add(password="Secret123")
    kdc.principal("host/myhost").add(password=None)
    client.local.user("tuser").add(password="Secret123")

    client.sssd.common.kcm(kdc)
    client.sssd.kcm["ccache_storage"] = ccache_storage
    client.sssd.start()

    with client.ssh("tuser", "Secret123") as ssh:
        with client.auth.kerberos(ssh) as krb:
            assert krb.cache_count() == 0, "KRB cache is not empty!"

            krb.kinit("alice", password="Secret123")
            assert krb.cache_count() == 1, "KRB cache value is not 1!"
            assert krb.has_primary_cache("alice", kdc.realm), "User 'alice' missing in cache!"
            assert krb.has_tickets("alice", kdc.realm, [kdc.tgt]), "No ticket for user 'alice' found!"

            krb.kinit("bob", password="Secret123")
            assert krb.cache_count() == 2, "KRB cache value is not 2!"
            assert krb.has_primary_cache("bob", kdc.realm), "User 'bob' missing in cache!"
            assert krb.has_tickets("bob", kdc.realm, [kdc.tgt]), "No ticket for user 'bob' found!"

            krb.kinit("carol", password="Secret123")
            assert krb.cache_count() == 3, "KRB cache value is not 3!"
            assert krb.has_primary_cache("carol", kdc.realm), "User 'carol' missing in cache!"
            assert krb.has_tickets("carol", kdc.realm, [kdc.tgt]), "No ticket for user 'carol' found!"

            krb.kvno("host/myhost")
            assert krb.cache_count() == 3, "KRB cache value is not 3!"
            assert krb.has_primary_cache("carol", kdc.realm), "User 'carol' missing in cache!"
            assert krb.has_tickets("alice", kdc.realm, [kdc.tgt]), "No ticket for user 'alice' found!"
            assert krb.has_tickets("bob", kdc.realm, [kdc.tgt]), "No ticket for user 'bob' found!"
            assert krb.has_tickets(
                "carol", kdc.realm, [kdc.tgt, kdc.qualify("host/myhost")]
            ), "No ticket for user 'carol' found!"

            # kdestroy 'carol' is the last primary cache
            krb.kdestroy()
            assert krb.cache_count() == 2, "KRB cache value is not 2!"

            # kinit 'carol' again
            krb.kinit("carol", password="Secret123")
            assert krb.cache_count() == 3, "KRB cache value is not 3!"
            assert krb.has_primary_cache("carol", kdc.realm), "User 'carol' missing in cache!"
            assert krb.has_tickets("alice", kdc.realm, [kdc.tgt]), "No ticket for user 'alice' found!"
            assert krb.has_tickets("bob", kdc.realm, [kdc.tgt]), "No ticket for user 'bob' found!"
            assert krb.has_tickets("carol", kdc.realm, [kdc.tgt]), "No ticket for user 'carol' found!"

            # kdestroy all
            krb.kdestroy(all=True)
            assert krb.cache_count() == 0


@pytest.mark.importance("critical")
@pytest.mark.authentication
@pytest.mark.topology(KnownTopology.Client)
@pytest.mark.parametrize("ccache_storage", ["memory", "secdb"])
def test_kcm__kswitch_between_primary_ccaches(client: Client, kdc: KDC, ccache_storage: str):
    """
    :title: Switching between primary ccaches.
    :setup:
        1. Add Kerberos principal "alice" to KDC
        2. Add Kerberos principal "bob" to KDC
        3. Add Kerberos service principal "host/alice" to KDC
        4. Add Kerberos service principal "host/bob" to KDC
        5. Add local user "tuser"
        6. Set ccache_storage in sssd.conf to @ccache_storage
        7. Start SSSD
    :steps:
        1. Authenticate as "tuser" over SSH
        2. Count existing credential caches
        3. Kinit as "alice"
        4. Kinit as "bob"
        5. Kswitch to "alice"
        6. Obtain ticket for "host/alice"
        7. Kswitch to "bob"
        8. Obtain ticket for "host/bob"
    :expectedresults:
        1. User is logged into the host
        2. Returns 0, no ccache is available
        3. TGT for "alice" was obtained, "alice" is the primary ccache
        4. TGT for "bob" was obtained, "bob" is the primary ccache
        5. "alice" is the primary ccache
        6. 2 ccaches exit, "alice" has TGT and "host/alice" tickets, "bob" has only TGT
        7. "bob" is the primary ccache
        8. 2 ccaches exit, "alice" has TGT and "host/alice" tickets, "bob" has TGT and "host/bob"
    :customerscenario: False
    """
    kdc.principal("alice").add(password="Secret123")
    kdc.principal("bob").add(password="Secret123")
    kdc.principal("host/alice").add(password=None)
    kdc.principal("host/bob").add(password=None)
    client.local.user("tuser").add(password="Secret123")

    client.sssd.common.kcm(kdc)
    client.sssd.kcm["ccache_storage"] = ccache_storage
    client.sssd.start()

    with client.ssh("tuser", "Secret123") as ssh:
        with client.auth.kerberos(ssh) as krb:
            assert krb.cache_count() == 0, "KRB cache is not empty!"

            krb.kinit("alice", password="Secret123")
            assert krb.has_primary_cache("alice", kdc.realm), "User 'alice' missing in cache!"

            krb.kinit("bob", password="Secret123")
            assert krb.has_primary_cache("bob", kdc.realm), "User 'bob' missing in cache!"

            krb.kswitch("alice", kdc.realm)
            assert krb.has_primary_cache("alice", kdc.realm), "User 'alice' missing in cache!"

            krb.kvno("host/alice")
            assert krb.cache_count() == 2, "KRB cache value is not 2!"
            assert krb.has_tickets(
                "alice", kdc.realm, [kdc.tgt, kdc.qualify("host/alice")]
            ), "No ticket for user 'alice' found!"
            assert krb.has_tickets("bob", kdc.realm, [kdc.tgt]), "No ticket for user 'bob' found!"

            krb.kswitch("bob", kdc.realm)
            krb.kvno("host/bob")
            assert krb.cache_count() == 2, "KRB cache value is not 2!"
            assert krb.has_tickets(
                "alice", kdc.realm, [kdc.tgt, kdc.qualify("host/alice")]
            ), "No ticket for user 'alice' found!"
            assert krb.has_tickets(
                "bob", kdc.realm, [kdc.tgt, kdc.qualify("host/bob")]
            ), "No ticket for user 'bob' found!"


@pytest.mark.importance("critical")
@pytest.mark.authentication
@pytest.mark.topology(KnownTopology.Client)
@pytest.mark.parametrize("ccache_storage", ["memory", "secdb"])
def test_kcm__subsidiary_ccaches_are_used_by_the_kcm(client: Client, kdc: KDC, ccache_storage: str):
    """
    :title: Subsidiary ccaches are usable and KCM: without UID can identify the collection.
    :setup:
        1. Add Kerberos principal "alice" to KDC
        2. Add Kerberos principal "bob" to KDC
        3. Add Kerberos service principal "host/alice" to KDC
        4. Add Kerberos service principal "host/bob" to KDC
        5. Add local user "tuser"
        6. Set ccache_storage in sssd.conf to @ccache_storage
        7. Start SSSD
    :steps:
        1. Authenticate as "tuser" over SSH
        2. Count existing credential caches
        3. Kinit as "alice"
        4. Obtain ticket for "host/alice"
        5. Kinit as "bob"
        6. Obtain ticket for "host/bob"
        7. Count existing credential caches
        8. Iterate over ccaches using KRB5CCNAME environment variable
        9. List all principals with KRB5CCNAME=KCM:
    :expectedresults:
        1. User is logged into the host
        2. Returns 0, no ccache is available
        3. Is successful
        4. Is successful
        5. Is successful
        6. Is successful
        7. Returns 2
        8. "alice" ccache contains TGT and "host/alice", "bob" ccache contains TGT and "host/bob"
        9. "alice" and "bob" ccaches are available
    :customerscenario: False
    """
    kdc.principal("alice").add(password="Secret123")
    kdc.principal("bob").add(password="Secret123")
    kdc.principal("host/alice").add(password=None)
    kdc.principal("host/bob").add(password=None)
    client.local.user("tuser").add(password="Secret123")

    client.sssd.common.kcm(kdc)
    client.sssd.kcm["ccache_storage"] = ccache_storage
    client.sssd.start()

    with client.ssh("tuser", "Secret123") as ssh:
        with client.auth.kerberos(ssh) as krb:
            assert krb.cache_count() == 0, "KRB cache value is not 0!"

            krb.kinit("alice", password="Secret123")
            krb.kvno("host/alice")

            krb.kinit("bob", password="Secret123")
            krb.kvno("host/bob")

            expected = {
                kdc.qualify("alice"): [kdc.tgt, kdc.qualify("host/alice")],
                kdc.qualify("bob"): [kdc.tgt, kdc.qualify("host/bob")],
            }

            assert krb.cache_count() == 2, "KRB cache value is not 2!"
            for principal, ccache in krb.list_ccaches().items():
                principals = krb.list_principals(env={"KRB5CCNAME": ccache})
                assert len(principals) == 1, "Principals count is not 1!"
                assert principal in principals, f"{principal} not in {principals}!"
                assert principals[principal] == expected[principal], "Principal ccache contains incorrect data!"

            principals = krb.list_principals(env={"KRB5CCNAME": "KCM:"})
            assert len(principals) == 2, "KCM principals count is not 2!"
            assert kdc.qualify("alice") in principals, "'alice' not in principals!"
            assert kdc.qualify("bob") in principals, "'bob' not in principals!"
            assert (
                principals[kdc.qualify("alice")] == expected[kdc.qualify("alice")]
            ), "Principal 'alice' in KCM does not match 'alice' in ccache!"
            assert (
                principals[kdc.qualify("bob")] == expected[kdc.qualify("bob")]
            ), "Principal 'bob' in KCM does not match 'bob' in ccache!"


@pytest.mark.importance("critical")
@pytest.mark.authentication
@pytest.mark.topology(KnownTopology.Client)
@pytest.mark.parametrize("ccache_storage", ["memory", "secdb"])
def test_kcm__kdestroy_nocache_throws_no_error(client: Client, kdc: KDC, ccache_storage: str):
    """
    :title: Destroying non-existing cache must not throw an error.
    :setup:
        1. Add local user "tuser"
        2. Set ccache_storage in sssd.conf to @ccache_storage
        3. Start SSSD
    :steps:
        1. Authenticate as "tuser" over SSH
        2. Count existing credential caches
        3. Run kdestroy
    :expectedresults:
        1. User is logged into the host
        2. Returns 0, no ccache is available
        3. Returns with return code 0
    :customerscenario: False
    """
    client.local.user("tuser").add(password="Secret123")

    client.sssd.common.kcm(kdc)
    client.sssd.kcm["ccache_storage"] = ccache_storage
    client.sssd.start()

    with client.ssh("tuser", "Secret123") as ssh:
        with client.auth.kerberos(ssh) as krb:
            assert krb.cache_count() == 0, "KRB cache value is not 0!"
            try:
                krb.kdestroy()
            except Exception as e:
                assert False, f"Destroying cache raised an error: {e}"


@pytest.mark.importance("critical")
@pytest.mark.authentication
@pytest.mark.topology(KnownTopology.Client)
def test_kcm__tgt_renewal_updates_ticket_as_configured(client: Client, kdc: KDC):
    """
    :title: Automatic ticket-granting ticket renewal.
    :setup:
        1. Add Kerberos principal "tuser" to KDC
        2. Add local user "tuser"
        3. Enable TGT renewal in KCM
        4. Start SSSD
    :steps:
        1. Authenticate as "tuser" over SSH
        2. Kinit as "tuser" and request renewable ticket
        3. Wait until automatic renewal is triggered and check that it was renewed
    :expectedresults:
        1. User is logged into the host
        2. TGT is available
        3. TGT was renewed
    :customerscenario: False
    """
    kdc.principal("tuser").add(password="Secret123")
    client.local.user("tuser").add(password="Secret123")

    client.sssd.common.kcm(kdc)
    client.sssd.kcm.update(
        tgt_renewal="True",
        krb5_renew_interval="1s",
    )
    client.sssd.start()

    with client.ssh("tuser", "Secret123") as ssh:
        with client.auth.kerberos(ssh) as krb:
            krb.kinit("tuser", password="Secret123", args=["-r", "2s", "-l", "2s"])
            (init_start, _) = krb.list_tgt_times(kdc.realm)
            time.sleep(2)
            (renew_start, _) = krb.list_tgt_times(kdc.realm)

            assert init_start < renew_start, "Initial renewal times exceeds renewal interval!"


@pytest.mark.topology(KnownTopology.Client)
def test_kcm__kinit_user_after_login(client: Client, kdc: KDC):
    """
    :title: kinit is successful after user login
    :setup:
        1. Add 'user1' to kdc and set its password
        2. Add 'user1' to local and set its password
        3. Configure Kerberos to allow KCM tests
    :steps:
        1. Authenticate user with ssh
        2. Authenticate to kerberos
        3. Call "kinit" with correct password
        4. Call "kinit" with wrong password
        5. Call "klist"
    :expectedresults:
        1. User is authenticated
        2. User is authenticated
        3. Call is successful
        4. Call is not successful
        5. Call is successful
    :customerscenario: False
    """
    username = "user1"
    password = "Secret123"

    kdc.principal(username).add(password=password)
    client.local.user(username).add(password=password)
    client.sssd.common.kcm(kdc)

    with client.ssh(username, password) as ssh:
        with client.auth.kerberos(ssh) as krb:
            assert krb.kinit(username, password=password).rc == 0, "kinit failed!"
            with pytest.raises(ProcessError):
                krb.kinit(username, password="wrong")
            assert krb.klist().rc == 0, "klist failed!"


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopology.Client)
def test_kcm__debug_log_enabled(client: Client, kdc: KDC):
    """
    :title: Kcm debug is enabled after sssd-kcm restart, when
     "debug_level" in kcm section is set to 9
    :setup:
        1. Add Kerberos principal "user1" to KDC
        2. Add local user "user1"
        3. Remove log files
        4. Set "debug_level" in kcm config section to "0"
        5. Remove kcm log files
        6. Start SSSD
    :steps:
        1. Try to produce some debug messages e.g. kdestroy
        2. Check that kcm debug messages were not generated
        3. Set "debug_level" in kcm config section to "9"
        4. Restart kcm
        5. Try to produce some debug messages e.g. kdestroy
        6. Check that kcm debug messages were generated
    :expectedresults:
        1. No messages were generated
        2. Log file did not get bigger
        3. Successfully set
        4. Successfully restarted
        5. Some messages were generated
        6. Log file did get bigger
    :customerscenario: False
    """

    def kcm_log_length() -> int:
        try:
            output = client.fs.wc(kcm_log_file, lines=True).stdout
            return int(output.split()[0])
        except ProcessError:
            return 0

    user = "user1"
    password = "Secret123"
    kcm_log_file = "/var/log/sssd/sssd_kcm.log"

    kdc.principal(user).add(password=password)
    client.local.user(user).add(password=password)
    client.sssd.common.kcm(kdc)

    client.sssd.kcm["debug_level"] = "0"
    client.sssd.config_apply()

    client.ssh(user, password).exec(["rm", "-f", kcm_log_file], raise_on_error=False)
    client.sssd.start()

    start_log_length = kcm_log_length()

    with client.ssh(user, password) as ssh:
        with client.auth.kerberos(ssh) as krb:
            krb.kdestroy()

    end_log_nodebug = kcm_log_length()
    assert start_log_length == end_log_nodebug, "Debug messages present!"

    client.sssd.kcm["debug_level"] = "9"
    client.sssd.config_apply()
    assert client.svc.restart("sssd-kcm").rc == 0, "KCM restart failed!"

    with client.ssh(user, password) as ssh:
        with client.auth.kerberos(ssh) as krb:
            krb.kdestroy()

    end_log_debug = kcm_log_length()
    assert start_log_length + 100 < end_log_debug, "Debug messages missing!"


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopology.LDAP)
def test_kcm__ssh_login_creates_kerberos_ticket(client: Client, ldap: LDAP, kdc: KDC):
    """
    :title: kcm: Verify ssh login is successful with kcm as default
    :setup:
        1. Add user and principal
        2. Set kerberos as default auth provider
        3. Start SSSD
    :steps:
        1. Authenticate as "user1" over SSH using kcm
    :expectedresults:
        1. Authenticated successfully
    :customerscenario: False
    """
    ldap.user("user1").add()
    kdc.principal("user1").add()

    client.sssd.common.krb5_auth(kdc)
    client.sssd.domain["krb5_realm"] = kdc.realm
    client.sssd.domain["krb5_server"] = kdc.host.hostname
    client.sssd.start()

    with client.ssh("user1", "Secret123") as ssh:
        with client.auth.kerberos(ssh) as krb:
            res = krb.klist()
            assert res.rc == 0, "klist failed!"


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopology.Client)
def test_kcm__configure_max_uid_ccaches_with_different_values(client: Client, kdc: KDC):
    """
    :title: "max_uid_ccaches" are enforced and limit only specific user
    :setup:
        1. Add local user "user0" and "user1"
        2. Add 66 Kerberos principals to KDC
        3. Start SSSD
    :steps:
        1. Authenticate as "user0" over SSH
        2. Set "max_uid_ccaches" to "1" and check its enforcement
        3. Remove "max_uid_ccaches" to use the default value
        4. Check the enforcement of quotas
        5. Set "max_uid_ccaches" to "65" and check its enforcement
        6. Kinit principal "user65" as "user1"
        7. Call kdestroy to destroy all caches as "user0"
        8. Kinit principal "user64" as "user0"
    :expectedresults:
        1. Authenticated successfully
        2. "max_uid_ccaches" are properly enforced
        3. Removed successfully
        4. "max_uid_ccaches" are properly enforced
        5. "max_uid_ccaches" are properly enforced
        6. Kinit is successful
        7. Kdestroy is successful
        8. Kinit is successful
    :customerscenario: False
    """
    user0 = "user0"
    user1 = "user1"
    password = "Secret123"
    client.local.user(user0).add(password=password)
    client.local.user(user1).add(password=password)

    for i in range(66):
        user = f"user{i}"
        kdc.principal(user).add(password=password)

    client.sssd.common.kcm(kdc)
    client.sssd.start()

    with client.ssh(user0, password) as ssh:
        with client.auth.kerberos(ssh) as krb:

            # max_uid_ccaches set to 1
            client.sssd.kcm["max_uid_ccaches"] = "1"
            client.sssd.config_apply()
            client.svc.restart("sssd-kcm")
            assert krb.kinit(user0, password=password).rc == 0, "max_uid_ccache = 1, kinit failed!"
            with pytest.raises(ProcessError):
                krb.kinit(user1, password=password)

            # max_uid_ccaches set to default (64)
            client.sssd.config.remove_option("kcm", "max_uid_ccaches")
            client.sssd.config_apply()
            client.svc.restart("sssd-kcm")
            for i in range(1, 64):
                user = f"user{i}"
                assert krb.kinit(user, password=password).rc == 0, "max_uid_ccache = 64, kinit failed!"
            with pytest.raises(ProcessError):
                krb.kinit("user64", password=password)

            # max_uid_ccaches set to 65
            client.sssd.kcm["max_uid_ccaches"] = "65"
            client.sssd.config_apply()
            client.svc.restart("sssd-kcm")
            assert krb.kinit("user64", password=password).rc == 0, "max_uid_ccache = 65, kinit failed!"
            with pytest.raises(ProcessError):
                krb.kinit("user65", password=password)

    # kinit as another user
    with client.ssh(user1, password) as ssh:
        with client.auth.kerberos(ssh) as krb:
            assert krb.kinit("user65", password=password).rc == 0, "kinit failed!"

    # kdestroy and then kinit
    with client.ssh("user0", password) as ssh:
        with client.auth.kerberos(ssh) as krb:
            assert krb.kdestroy(all=True).rc == 0, "kdestroy all tickets failed!"
            assert krb.kinit("user65", password=password).rc == 0, "kinit failed!"
